from dbnd._core.commands.metrics import (
    log_artifact,
    log_dataframe,
    log_duration,
    log_metric,
    log_metrics,
)
from dbnd._core.plugin.dbnd_plugins import assert_plugin_enabled


def get_spark_session():
    assert_plugin_enabled("dbnd-spark")
    import dbnd_spark

    return dbnd_spark.get_spark_session()



from dbnd._core.constants import TaskRunState
from dbnd._core.parameter.parameter_definition import ParameterGroup, _ParameterKind
from dbnd._core.tracking.schemas.base import ApiObjectSchema
from dbnd._core.tracking.schemas.tracking_info_objects import (
    TaskDefinitionInfo,
    TaskRunInfo,
    TaskRunParamInfo,
)
from dbnd._core.utils.dotdict import _as_dotted_dict
from dbnd._vendor.marshmallow import fields, post_load
from dbnd._vendor.marshmallow_enum import EnumField


class TaskDefinitionParamSchema(ApiObjectSchema):
    """
    Based on TaskDefinitionParam object
    """

    name = fields.String()

    default = fields.String(allow_none=True)

    description = fields.String()

    group = EnumField(ParameterGroup)
    kind = EnumField(_ParameterKind)
    load_on_build = fields.Boolean()

    significant = fields.Boolean()
    value_type = fields.String()

    @post_load
    def make_task_definition_param(self, data, **kwargs):
        return _as_dotted_dict(**data)


class TaskDefinitionInfoSchema(ApiObjectSchema):
    task_definition_uid = fields.UUID()
    name = fields.String()

    class_version = fields.String()
    family = fields.String()

    module_source = fields.String(allow_none=True)
    module_source_hash = fields.String(allow_none=True)

    source = fields.String(allow_none=True)
    source_hash = fields.String(allow_none=True)

    type = fields.String()

    task_param_definitions = fields.Nested(TaskDefinitionParamSchema, many=True)

    @post_load
    def make_task_definition(self, data, **kwargs):
        return TaskDefinitionInfo(**data)


class TaskRunParamSchema(ApiObjectSchema):
    parameter_name = fields.String()
    value_origin = fields.String()
    value = fields.String()

    @post_load
    def make_task_run_param(self, data, **kwargs):
        return TaskRunParamInfo(**data)


class TaskRunInfoSchema(ApiObjectSchema):
    task_run_uid = fields.UUID()
    task_run_attempt_uid = fields.UUID()

    task_definition_uid = fields.UUID()
    run_uid = fields.UUID()
    task_id = fields.String()
    task_signature = fields.String()
    task_signature_source = fields.String()

    task_af_id = fields.String()
    execution_date = fields.DateTime()

    name = fields.String()

    env = fields.String()

    command_line = fields.String()
    functional_call = fields.String()

    has_downstreams = fields.Boolean()
    has_upstreams = fields.Boolean()

    is_reused = fields.Boolean()
    is_dynamic = fields.Boolean()
    is_system = fields.Boolean()
    is_skipped = fields.Boolean()
    is_root = fields.Boolean()
    output_signature = fields.String()

    state = EnumField(TaskRunState)
    target_date = fields.Date(allow_none=True)

    log_local = fields.String(allow_none=True)
    log_remote = fields.String(allow_none=True)

    version = fields.String()

    task_run_params = fields.Nested(TaskRunParamSchema, many=True)

    external_links = fields.Dict(allow_none=True)

    @post_load
    def make_task_run(self, data, **kwargs):
        return TaskRunInfo(**data)



from __future__ import print_function

import functools
import logging
import os

import attr
import six

from dbnd._core.cli.click_utils import ConfigValueType, _help
from dbnd._core.cli.service_auto_completer import completer
from dbnd._core.configuration.config_readers import parse_and_build_config_store
from dbnd._core.configuration.environ_config import (
    ENV_DBND__TRACKING,
    new_dbnd_project_config_context,
)
from dbnd._core.configuration.pprint_config import pformat_config_store_as_table
from dbnd._core.context.bootstrap import dbnd_bootstrap
from dbnd._core.log.config import configure_basic_logging
from dbnd._core.task_build.task_metaclass import TaskMetaclass
from dbnd._core.task_build.task_registry import get_task_registry
from dbnd._core.tracking.schemas.tracking_info_run import ScheduledRunInfo
from dbnd._core.utils.basics.dict_utils import filter_dict_remove_false_values
from dbnd._core.utils.basics.environ_utils import env as env_context
from dbnd._vendor import click
from dbnd._vendor.click_tzdatetime import TZAwareDateTime


logger = logging.getLogger(__name__)


def _to_conf(kwargs):
    return {k: str(v) for k, v in kwargs.items() if v is not None}


def build_dynamic_task(original_cls, new_cls_name):
    # Helper func to allow using --type-name feature
    original_name = original_cls.task_definition.task_family
    logger.info("Creating new class %s from %s", new_cls_name, original_name)
    attributes = {}  # {'_conf__task_family': original_cls.get_task_family()}
    if original_cls._conf__decorator_spec:
        conf__decorator_spec = attr.evolve(
            original_cls._conf__decorator_spec, name=new_cls_name
        )
        attributes["_conf__decorator_spec"] = conf__decorator_spec
    else:
        attributes["_conf__task_family"] = new_cls_name
    new_cls = TaskMetaclass(str(new_cls_name), (original_cls,), attributes)
    return new_cls


@click.command(context_settings=dict(help_option_names=[]))
@click.argument("task", autocompletion=completer.task(), required=False)
@click.option(
    "--module",
    "-m",
    help="Define a path of a module where DBND can search for a task or pipeline",
)
@click.option(
    "--set",
    "-s",
    "_sets",
    help="Set a configuration value (task_name.task_parameter=value)",
    type=ConfigValueType(),
    multiple=True,
    autocompletion=completer.task_param(),
)
@click.option(
    "--set-config",
    "-c",
    "_sets_config",
    help="Set configuration a value (key=value). Example: --set run.name=my_run",
    type=ConfigValueType(),
    multiple=True,
    autocompletion=completer.config_param(),
)
@click.option(
    "--set-root",
    "-r",
    "_sets_root",
    help="Set or override a task parameter value (task_parameter=value)",
    type=ConfigValueType(),
    multiple=True,
    autocompletion=completer.root_param(),
)
@click.option(
    "--verbose", "-v", count=True, help="Make the logging output more verbose"
)
@click.option(
    "--print-task-band", is_flag=True, help="Print task_band in the logging output."
)
@click.option("--describe", is_flag=True, help="Describe the current run")
@click.option(
    "--env",
    default="local",
    show_default=True,
    help="task environment: local/aws/aws_prod/gcp/prod",
)
@click.option("--parallel", is_flag=True, help="Run tasks in parallel")
@click.option("--conf-file", help="List of files to read from")
@click.option("--task-version", help="Task version directly affects the task signature")
@click.option("--project-name", help="Name of this Databand project")
@click.option("--name", help="Name of this run")
@click.option("--description", help="Description of this run")
@click.option(
    "--override",
    "-o",
    "_overrides",
    type=ConfigValueType(),
    multiple=True,
    help="Override configuration value (higher priority than any config source) key=value",
)
@click.option(
    "--help", "is_help", is_flag=True, help="Used for dynamic loading of modules"
)
@click.option("--run-driver", "run_driver", help="Running in remote mode")
@click.option("--task-name", "alternative_task_name", help="Name of this task")
@click.option(
    "--scheduled-job-name",
    "-sjn",
    help="Associate this run with the scheduled job which will be created if needed",
)
@click.option(
    "--scheduled-date",
    "-sd",
    help="Should be used only when scheduled-job-name is set",
    type=TZAwareDateTime(),
)
@click.option("--interactive", is_flag=True, help="Run submission in blocking mode")
@click.option(
    "--submit-driver",
    "submit_driver",
    flag_value=1,
    help="Run driver at a remote engine",
)
@click.option(
    "--local-driver", "submit_driver", flag_value=0, help="Run the driver locally"
)
@click.option(
    "--submit-tasks",
    "submit_tasks",
    flag_value=1,
    help="Submit tasks from driver to a remote engine",
)
@click.option(
    "--no-submit-tasks",
    "submit_tasks",
    flag_value=0,
    help="Disable task submission, run tasks in the driver process.",
)
@click.option(
    "--disable-web-tracker", help="Disable web tracking", is_flag=True,
)
@click.option("--interactive", is_flag=True, help="Run submission in blocking mode")
@click.option(
    "--open-web-tab", help="Open the tracker URL in the Databand UI", is_flag=True
)
@click.option(
    "--docker-build-tag",
    help="Define custom docker image tag for docker image that will be built",
)
@click.pass_context
def run(
    ctx,
    is_help,
    task,
    module,
    _sets,
    _sets_config,
    _sets_root,
    _overrides,
    verbose,
    print_task_band,
    describe,
    env,
    parallel,
    conf_file,
    task_version,
    project_name,
    name,
    description,
    run_driver,
    alternative_task_name,
    scheduled_job_name,
    scheduled_date,
    interactive,
    submit_driver,
    submit_tasks,
    disable_web_tracker,
    open_web_tab,
    docker_build_tag,
):
    """
    Run a task or a DAG

    To see tasks use `dbnd show-tasks` (tab completion is available).
    """

    from dbnd._core.context.databand_context import new_dbnd_context, DatabandContext
    from dbnd._core.utils.structures import combine_mappings
    from dbnd import config

    task_name = task
    # --verbose, --describe, --env, --parallel, --conf-file and --project-name
    # we filter out false flags since otherwise they will always override the config with their falseness
    main_switches = dict(
        databand=filter_dict_remove_false_values(
            dict(
                verbose=verbose > 0,
                task_band=print_task_band,
                describe=describe,
                env=env,
                conf_file=conf_file,
                project_name=project_name,
            )
        ),
        run=filter_dict_remove_false_values(
            dict(
                name=name,
                parallel=parallel,
                description=description,
                is_archived=describe,
            )
        ),
    )

    if submit_driver is not None:
        main_switches["run"]["submit_driver"] = bool(submit_driver)
    if submit_tasks is not None:
        main_switches["run"]["submit_tasks"] = bool(submit_tasks)
    if disable_web_tracker:
        main_switches.setdefault("core", {})["tracker_api"] = "disabled"

    if task_version is not None:
        main_switches["task"] = {"task_version": task_version}

    cmd_line_config = parse_and_build_config_store(
        source="cli", config_values=main_switches
    )

    _sets = list(_sets)
    _sets_config = list(_sets_config)
    _sets_root = list(_sets_root)

    root_task_config = {}
    for _set in _sets_root:
        root_task_config = combine_mappings(left=root_task_config, right=_set)

    # remove all "first level" config values, assume that they are for the main task
    # add them to _sets_root
    for _set in _sets:
        for k, v in list(_set.items()):
            # so json-like values won't be included
            if "." not in k and isinstance(v, six.string_types):
                root_task_config[k] = v
                del _set[k]

    # --set, --set-config
    if _sets:
        cmd_line_config.update(_parse_cli(_sets, source="--set"))
    if _sets_config:
        cmd_line_config.update(_parse_cli(_sets_config, source="--set-config"))
    if _overrides:
        cmd_line_config.update(
            _parse_cli(_overrides, source="--set-override", override=True)
        )
    if interactive:
        cmd_line_config.update(
            _parse_cli([{"run.interactive": True}], source="--interactive")
        )
    if verbose > 1:
        cmd_line_config.update(
            _parse_cli([{"task_build.verbose": True}], source="-v -v")
        )

    if cmd_line_config:
        config.set_values(cmd_line_config, source="cmdline")
    if verbose:
        logger.info("CLI config: \n%s", pformat_config_store_as_table(cmd_line_config))

    # double checking on bootstrap, as we can run from all kind of locations
    # usually we should be bootstraped already as we run from cli.
    dbnd_bootstrap()
    if not config.getboolean("log", "disabled"):
        configure_basic_logging(None)

    scheduled_run_info = None
    if scheduled_job_name:
        scheduled_run_info = ScheduledRunInfo(
            scheduled_job_name=scheduled_job_name, scheduled_date=scheduled_date
        )

    with new_dbnd_project_config_context(
        **{ENV_DBND__TRACKING: "False"}
    ), new_dbnd_context(
        name="run", module=module
    ) as context:  # type: DatabandContext
        task_registry = get_task_registry()

        tasks = task_registry.list_dbnd_task_classes()
        completer.refresh(tasks)

        # modules are loaded, we can load the task
        task_cls = None
        if task_name:
            task_cls = task_registry.get_task_cls(task_name)
            if alternative_task_name:
                task_cls = build_dynamic_task(
                    original_cls=task_cls, new_cls_name=alternative_task_name
                )
                task_name = alternative_task_name

        # --set-root
        # now we can get it config, as it's not main task, we can load config after the configuration is loaded
        if task_cls is not None and root_task_config:
            # adding root task to configuration
            config.set_values(
                {task_cls.task_definition.task_config_section: root_task_config},
                source="--set-root",
            )

        if is_help or not task_name:
            print_help(ctx, task_cls)
            return

        if open_web_tab:
            config.set_values({"run": {"open_web_tracker_in_browser": "True"}})

        if docker_build_tag:
            config.set_values({"kubernetes": {"docker_build_tag": docker_build_tag}})

        return context.dbnd_run_task(
            task_or_task_name=task_name,
            run_uid=run_driver,
            scheduled_run_info=scheduled_run_info,
        )


def is_running_in_direct_db_mode(ctx):
    return "api" in ctx.settings.core.tracker and ctx.settings.core.tracker_api == "db"


def print_help(ctx, task_cls):
    from dbnd._core.parameter.parameter_definition import _ParameterKind

    formatter = ctx.make_formatter()
    run.format_help(ctx, formatter)
    if task_cls:
        dl = []
        for (param_name, param_obj) in task_cls.task_definition.all_task_params.items():
            if param_obj.system or param_obj.kind == _ParameterKind.task_output:
                continue

            param_help = _help(param_obj.description)
            dl.append(("-r " + param_name, param_help))

        with formatter.section("Task"):
            formatter.write_dl(dl)
    click.echo(formatter.getvalue().rstrip("\n"), color=ctx.color)


def _parse_cli(configs, source, override=False):
    """
    Parse every item in configs , joining them into one big ConfigStore
    """
    config_values_list = [
        parse_and_build_config_store(
            config_values=c, source=source, override=override, auto_section_parse=True
        )
        for c in configs
    ]
    return functools.reduce((lambda x, y: x.update(y)), config_values_list)



from dbnd._core.constants import _ConfigParamContainer
from dbnd._core.errors.base import ConfigLookupError
from dbnd._core.task.base_task import _BaseTask
from dbnd._core.utils.basics.singleton_context import SingletonContext


class Config(_BaseTask, SingletonContext, _ConfigParamContainer):
    """
    Class for configuration. See :ref:`ConfigClasses`.
    The nice thing is that we can instantiate this class
    and get an object with all the environment variables set.
    This is arguably a bit of a hack.
    """

    def __str__(self):
        return self.task_name

    def params_to_env_map(self):
        self._params

    @classmethod
    def current(cls, name=None):
        """
        Syntax sugar for accessing the current config instance.
        """
        from dbnd._core.current import get_databand_context

        if not name:
            if cls._conf__task_family:
                # using the current cls section name to get the current instance of a class
                name = cls._conf__task_family
            else:
                raise ConfigLookupError(
                    "name is required for retrieving a config instance"
                )

        return get_databand_context().settings.get_config(name)


Config.task_definition.hidden = True

config_singletons = {}



import contextlib
import functools
import logging
import os
import sys
import typing

from typing import Any, Mapping, Optional, Union

import attr

from dbnd._core.configuration.config_readers import (
    _default_configuration_paths,
    parse_and_build_config_store,
    read_environ_config,
    read_from_config_files,
)
from dbnd._core.configuration.config_store import (
    _ConfigMergeSettings,
    _ConfigStore,
    _lower_config_name,
)
from dbnd._core.configuration.config_value import ConfigValue
from dbnd._core.configuration.pprint_config import (
    pformat_all_layers,
    pformat_current_config,
)
from dbnd._core.context.bootstrap import dbnd_system_bootstrap
from dbnd._core.utils.basics.helpers import parse_bool
from dbnd._vendor.snippets.airflow_configuration import expand_env_var
from targets import target
from targets.values.builtins_values import STR_VALUE_TYPE


if typing.TYPE_CHECKING:
    from dbnd._core.configuration.config_store import _TConfigStore
    from dbnd._core.parameter.parameter_definition import ParameterDefinition

    from targets import Target

logger = logging.getLogger(__name__)


@attr.s(str=False)
class _ConfigLayer(object):
    name = attr.ib()  # type: str
    config = attr.ib()  # type: _TConfigStore
    layer_config = attr.ib()  # type: _TConfigStore
    parent = attr.ib(default=None)  # type: Optional[_ConfigLayer]

    def get_config_value(self, section, key):
        # type: (str, str)->Optional[ConfigValue]
        section = self.config.get(_lower_config_name(section))
        if not section:
            return None
        return section.get(_lower_config_name(key))

    def create_layer(
        self,
        name,  # type: str
        config_values,  # type: _TConfigStore
        merge_settings=None,  # type: _ConfigMergeSettings
    ):
        new_config = self.config.merge(
            config_values=config_values, merge_settings=merge_settings
        )
        return _ConfigLayer(
            name=name, parent=self, layer_config=config_values, config=new_config
        )

    def get_all_layers(self):
        trail = []
        current = self
        while current:
            trail.append(current)
            current = current.parent
        return trail

    @property
    def config_layer_path(self):
        return " <- ".join(l.name for l in self.get_all_layers())

    def __str__(self):
        return self.config_layer_path


@attr.s(str=False)
class DbndConfig(object):
    config_layer = attr.ib()  # type: _ConfigLayer
    # we need to initialize config with environment and config files first
    # if user already has config() calls , we will need to reapply his changes
    # on top of file and env configuration
    # User can run his config() before the first DatabandContext is called,
    # that means, we'll have to "replay his changes" on top of all this changes
    initialized_with_env = attr.ib(default=False)

    def _new_config_layer(
        self, config_values, source=None, override=False, merge_settings=None
    ):
        # let validate that we are initialized
        # user can call this function out of no-where, so we will create a layer, and will override it
        # the moment we create more layers on config.system_load
        dbnd_system_bootstrap()

        if not config_values:
            return self.config_layer
        if not isinstance(config_values, _ConfigStore):
            if not source:
                source = "{sig}".format(sig=id(config_values))
            config_values = parse_and_build_config_store(
                config_values=config_values, source=source, override=override
            )  # type: _ConfigStore

        source = source or config_values.source
        if not source:
            source = "{sig}".format(sig=id(config_values))
        return self.config_layer.create_layer(
            name=source, config_values=config_values, merge_settings=merge_settings
        )

    @contextlib.contextmanager
    def __call__(self, config_values=None, source=None, merge_settings=None):
        new_layer = self._new_config_layer(
            config_values, source=source, merge_settings=merge_settings
        )
        with self.config_layer_context(config_layer=new_layer):
            yield self

    @contextlib.contextmanager
    def config_layer_context(self, config_layer):
        current_layer = self.config_layer
        current_initialized_with_env = self.initialized_with_env
        # this will create new layer
        try:
            self.config_layer = config_layer
            yield self
        finally:
            self.config_layer = current_layer
            self.initialized_with_env = current_initialized_with_env

    def set_values(
        self, config_values, source=None, override=False, merge_settings=None
    ):
        # type: (Union[ Mapping[str, Mapping[str, Any]], _ConfigStore], str, bool ,_ConfigMergeSettings)-> None
        """
        Global override, changing current layout
        """
        self.config_layer = self._new_config_layer(
            config_values,
            source=source,
            merge_settings=merge_settings,
            override=override,
        )
        return self.config_layer

    def set_from_config_file(self, config_path):
        # type: (Union[Target,str]) -> None
        """
        Load Configuration from config file
        """
        config_path = target(config_path)
        config_values = read_from_config_files([config_path])
        self.set_values(
            config_values=config_values, source=os.path.basename(str(config_path))
        )

    def set(
        self, section, key, value, override=False, source=None, merge_settings=None
    ):
        self.set_values(
            {section: {key: value}},
            source=source,
            override=override,
            merge_settings=merge_settings,
        )

    def set_parameter(self, parameter, value, source=None):
        # type: (ParameterDefinition, object, str) -> None
        return self.set(parameter.task_family, parameter.name, value, source=source)

    def get_config_value(self, section, key):
        # type: (str, str)->Optional[ConfigValue]
        """
        Gets the value of the section/option using method.
        This is the function used by
        Returns default if value is not found.
        Raises an exception if the default value is not None and doesn't match the expected_type.
        """
        return self.config_layer.config.get_config_value(section, key)

    def get(self, section, key, default=None, expand_env=True):

        config_value = self.get_config_value(section=section, key=key)
        if config_value:
            value = config_value.value
            if expand_env:
                value = expand_env_var(value)
                value = STR_VALUE_TYPE._interpolate_from_str(value)
            return value
        return default

    def getboolean(self, section, key, **kwargs):
        return parse_bool(self.get(section, key, **kwargs))

    def getint(self, section, key, **kwargs):
        return int(self.get(section, key, **kwargs))

    def getfloat(self, section, key, **kwargs):
        return float(self.get(section, key, **kwargs))

    def load_system_configs(self, force=False):
        if self.initialized_with_env and not force:
            return

        # first, let read from files
        system_config_store = read_from_config_files(_default_configuration_paths())

        # now we can merge with environment values
        system_config_store.update(read_environ_config())

        # all configs will be added as one layer called 'system'
        self.set_values(system_config_store, source="system")
        self.initialized_with_env = True

    def __str__(self):
        return "Config[%s]" % self.config_layer.name

    def log_current_config(self, sections=None, as_table=False):
        logger.info(pformat_current_config(self, sections=sections, as_table=as_table))

    def log_layers(self, sections=None):
        logger.info(pformat_all_layers(self, sections=sections))

    @classmethod
    def build_empty(cls, name):
        return cls(
            config_layer=_ConfigLayer(
                name=name,
                config=_ConfigStore(),
                layer_config=_ConfigStore(),
                parent=None,
            )
        )


config = DbndConfig.build_empty(name="empty")


def config_deco(config_values, **deco_kwargs):
    """
    do not use it in PY3, use config instead
    function decorator for PY2
    """

    def decorator(method):
        @functools.wraps(method)
        def f(*args, **kwargs):
            with config(config_values, **deco_kwargs):
                return method(*args, **kwargs)

        return f

    return decorator



import typing


if typing.TYPE_CHECKING:
    from dbnd._core.run.databand_run import DatabandRun


class RunCtrl(object):
    def __init__(self, run):
        # type: (DatabandRun)-> None
        super(RunCtrl, self).__init__()

        self.run = run

    @property
    def context(self):
        return self.run.context

    @property
    def settings(self):
        return self.context.settings



import logging
import subprocess
import sys

from os import environ

from dbnd._core.configuration.environ_config import (
    ENV_DBND__ENV_IMAGE,
    ENV_DBND__ENV_MACHINE,
)
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import config
from dbnd._core.tracking.schemas.tracking_info_objects import TaskRunEnvInfo
from dbnd._core.utils.basics.text_banner import safe_string
from dbnd._core.utils.platform.windows_compatible.getuser import dbnd_getuser
from dbnd._core.utils.project.project_fs import project_path
from dbnd._core.utils.timezone import utcnow
from dbnd._core.utils.uid_utils import get_uuid
from targets.values.version_value import VersionStr


class RunInfoConfig(config.Config):
    """(Advanced) Databand's run information gatherer"""

    _conf__task_family = "run_info"
    source_version = parameter(
        default="git", description="gather version control via git/None"
    )[VersionStr]
    user_data = parameter(default=None, description="UserData")[str]
    user = parameter(default=None, description="override user name with the value")[str]

    def build_task_run_info(self):
        task_run_env_uid = get_uuid()
        import dbnd

        logging.debug("Created new task run env with uid '%s'", task_run_env_uid)

        machine = environ.get(ENV_DBND__ENV_MACHINE, "")
        if environ.get(ENV_DBND__ENV_IMAGE, None):
            machine += " image=%s" % environ.get(ENV_DBND__ENV_IMAGE)
        return TaskRunEnvInfo(
            uid=task_run_env_uid,
            databand_version=dbnd.__version__,
            user_code_version=self.source_version,
            user_code_committed=True,
            cmd_line=subprocess.list2cmdline(sys.argv),
            user=self.user or dbnd_getuser(),
            machine=machine,
            project_root=project_path(),
            user_data=safe_string(self.user_data, max_value_len=500),
            heartbeat=utcnow(),
        )



from targets.target_config import TargetConfig, parse_target_config
from targets.values import ValueType


class TargetConfigValueType(ValueType):  # extracted for generics support
    """
      ValueType whose value is a ```target```.

      In the task definition, use

      .. code-block:: python

          class MyTask(dbnd.Task):
              some_input = TargetValueType()

              def run(self):
                    self.some_input.open('r')

      """

    type = TargetConfig

    support_from_str = True

    def parse_from_str(self, x):
        return parse_target_config(x)



import logging
import typing

from typing import Any, Union

from dbnd._core.current import current_task_run, get_databand_run, is_verbose
from dbnd._core.decorator.task_decorator_spec import args_to_kwargs
from dbnd._core.errors import MissingParameterError
from targets.inline_target import InlineTarget


logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from dbnd._core.decorator.func_task_call import FuncCall
    from dbnd._core.task.task import Task


def create_dynamic_task(func_call):
    # type: (FuncCall) -> Task
    task_cls, call_args, call_kwargs = (
        func_call.task_cls,
        func_call.call_args,
        func_call.call_kwargs.copy(),
    )
    from dbnd import pipeline, PipelineTask
    from dbnd._core.decorator.dbnd_decorator import _default_output

    parent_task = current_task_run().task
    dbnd_run = get_databand_run()

    if task_cls._conf__decorator_spec is not None:
        # orig_call_args, orig_call_kwargs = call_args, call_kwargs
        call_args, call_kwargs = args_to_kwargs(
            task_cls._conf__decorator_spec.args, call_args, call_kwargs
        )

    # Map all kwargs to the "original" target of that objects
    # for example: for DataFrame we'll try to find a relevant target that were used to read it
    # get all possible value's targets
    call_kwargs_as_targets = dbnd_run.target_origin.get_for_map(call_kwargs)
    for p_name, value_origin in call_kwargs_as_targets.items():
        root_target = value_origin.origin_target
        path = root_target.path if hasattr(root_target, "path") else None
        call_kwargs[p_name] = InlineTarget(
            root_target=root_target,
            obj=call_kwargs[p_name],
            value_type=value_origin.value_type,
            source=value_origin.origin_target.source,
            path=path,
        )

    call_kwargs.setdefault("task_is_dynamic", True)
    call_kwargs.setdefault(
        "task_in_memory_outputs", parent_task.settings.dynamic_task.in_memory_outputs
    )

    # in case of pipeline - we'd like to run it as regular task
    # if False and issubclass(task_cls, PipelineTask):
    #     # TODO: do we want to support this behavior
    #     task_cls = task(task_cls._conf__decorator_spec.item).task_cls

    if issubclass(task_cls, PipelineTask):
        # if it's pipeline - create new databand run
        # create override _task_default_result to be object instead of target
        task_cls = pipeline(
            task_cls._conf__decorator_spec.item, _task_default_result=_default_output
        ).task_cls

        # instantiate inline pipeline
        t = task_cls(*call_args, **call_kwargs)
        return t
    else:
        # instantiate inline task
        t = task_cls(*call_args, **call_kwargs)

        # update upstream/downstream relations - needed for correct tracking
        # we can have the task as upstream , as it was executed already
        if not parent_task.task_dag.has_upstream(t):
            parent_task.set_upstream(t)
        return t


def run_dynamic_task_safe(task, func_call):
    # type: (Task, FuncCall) -> Union[Any]
    try:
        from dbnd._core.decorator.func_task_call import TaskCallState, CALL_FAILURE_OBJ

        task._dbnd_call_state = TaskCallState(should_store_result=True)
        # this is the real run of the decorated function
        from dbnd import PipelineTask

        dbnd_run = get_databand_run()

        if isinstance(task, PipelineTask):
            # if it's pipeline - create new databand run
            run = dbnd_run.context.dbnd_run_task(task)
            task_run = run.get_task_run(task.task_id)
        else:
            task_run = dbnd_run.run_dynamic_task(
                task, task_engine=current_task_run().task_engine
            )
            if task._dbnd_call_state.result_saved:
                return task._dbnd_call_state.result

        t = task_run.task
        # if we are inside run, we want to have real values, not deferred!
        if t.task_definition.single_result_output:
            return t.__class__.result.load_from_target(t.result)
            # we have func without result, just fallback to None

        return t
    except Exception:
        if task and task._dbnd_call_state:
            if task._dbnd_call_state.finished:
                # if function was invoked and finished - than we failed in dbnd post-exec
                # just return invoke_result to user
                _handle_dynamic_error("task-post-execute", func_call)
                return task._dbnd_call_state.result
            if task._dbnd_call_state.started:
                # if started but not finished -> it was user code exception -> re-raise
                raise

        # not started - our exception on pre-exec, run user code
        _handle_dynamic_error("task-pre-execute", func_call)
        return func_call.invoke()
    finally:
        # we'd better clean _invoke_result to avoid memory leaks
        task._dbnd_call_state = None


def create_and_run_dynamic_task_safe(func_call):
    try:
        task = create_dynamic_task(func_call)
    except MissingParameterError:
        # We can't handle MissingParameterError, function invocation will always fail
        raise
    except Exception:
        _handle_dynamic_error("task-create", func_call)
        return func_call.invoke()

    return run_dynamic_task_safe(task=task, func_call=func_call)


def _handle_dynamic_error(msg, func_call):
    if is_verbose():
        logger.warning(
            "Failed during dbnd %s for %s, ignoring, and continue without tracking/orchestration"
            % (msg, func_call.task_cls),
            exc_info=True,
        )
    else:
        logger.info(
            "Failed during dbnd %s for %s, ignoring, and continue without tracking"
            % (msg, func_call.task_cls)
        )



import abc

from dbnd._core.constants import TaskType
from dbnd._core.task.data_task import Task


class PipelineTask(Task):
    """
    Use for tasks that only wrap other tasks and that by definition are done if all their requirements exist.
    """

    _conf__task_type_name = TaskType.pipeline

    def _task_run(self):
        result = self.run()
        return result

    @abc.abstractmethod
    def band(self):
        """
            This is the method you should override while using PipelineTask.
            we need to implement result of this function to be "output" of the task (task_output, and vars)
            Your Pipeline.band() call should have one or more tasks wired one into another.
            See examples!"
        :return:
        """
        return

    def _complete(self):
        if self.task_band:
            if not self.task_band.exists():
                return False
            # With very large pipelines, checking all tasks might take a very long time
            # so we might want to assume that if the band exist, probably all outputs also exist
            if self.settings.run.pipeline_band_only_check:
                return True
        return super(PipelineTask, self)._complete()


PipelineTask.task_definition.hidden = True



import os
import re

from dbnd._core.errors import DatabandError
from targets.file_target import FileTarget
from targets.fs.local import LocalFileSystem
from targets.multi_target import MultiTarget
from targets.target_config import extract_target_config_from_path
from targets.utils.path import trailing_slash


_GLOB_PATH_REGEX = re.compile(r".*{.*,.*}.*")


def _is_glob_path(path):
    match = _GLOB_PATH_REGEX.match(path)
    return match is not None


def target(*path, **kwargs):
    """
    autoresolving function
    :param path:
    :param kwargs:
    :return: FileTarget
    """

    path = [str(p) for p in path]
    path = os.path.join(*path)
    if not path:
        raise DatabandError("Can not convert empty string '%s' to Target" % path)
    if not _is_glob_path(path):
        if "," in path:
            return MultiTarget(targets=[target(p, **kwargs) for p in path.split(",")])

    fs = kwargs.pop("fs", None)
    config = kwargs.pop("config", None)
    config = extract_target_config_from_path(path, config=config)
    if path.endswith("[noflag]"):
        config = config.with_flag(None)
        path = path[:-8]

    if config.folder and not trailing_slash(path):
        path = "%s%s" % (path, os.path.sep if isinstance(fs, LocalFileSystem) else "/")

    if config.target_factory:
        return config.target_factory(path, fs=fs, config=config)

    if config.folder:
        from targets.dir_target import DirTarget

        return DirTarget(path, fs=fs, config=config)
    return FileTarget(path=path, fs=fs, config=config, **kwargs)



import logging
import typing

from dbnd._core.constants import TaskExecutorType
from dbnd._core.errors import DatabandConfigError, friendly_error
from dbnd._core.plugin.dbnd_plugins import is_airflow_enabled, is_plugin_enabled
from dbnd._core.task_executor.local_task_executor import LocalTaskExecutor


if typing.TYPE_CHECKING:
    from typing import List
    from dbnd._core.run.databand_run import DatabandRun
    from dbnd._core.settings import EngineConfig
    from dbnd._core.task_executor.task_executor import TaskExecutor
    from dbnd._core.task_run.task_run import TaskRun

logger = logging.getLogger(__name__)


def calculate_task_executor_type(submit_tasks, remote_engine, settings):
    run_config = settings.run
    parallel = run_config.parallel
    task_executor_type = run_config.task_executor_type

    if task_executor_type is None:
        if is_airflow_enabled():
            from dbnd_airflow.executors import AirflowTaskExecutorType

            task_executor_type = AirflowTaskExecutorType.airflow_inprocess
        else:
            task_executor_type = TaskExecutorType.local

    if is_airflow_enabled():
        from dbnd_airflow.executors import AirflowTaskExecutorType

        if parallel:
            if task_executor_type == TaskExecutorType.local:
                logger.warning(
                    "Auto switching to engine type '%s' due to parallel mode.",
                    AirflowTaskExecutorType.airflow_multiprocess_local,
                )
                task_executor_type = AirflowTaskExecutorType.airflow_multiprocess_local

            if task_executor_type == AirflowTaskExecutorType.airflow_inprocess:
                logger.warning(
                    "Auto switching to engine type '%s' due to parallel mode.",
                    AirflowTaskExecutorType.airflow_multiprocess_local,
                )
                task_executor_type = AirflowTaskExecutorType.airflow_multiprocess_local

        if (
            task_executor_type == AirflowTaskExecutorType.airflow_multiprocess_local
            or task_executor_type == AirflowTaskExecutorType.airflow_kubernetes
        ):
            from dbnd_airflow.db_utils import (
                airflow_sql_conn_url,
                airlow_sql_alchemy_conn,
            )

            if "sqlite" in airlow_sql_alchemy_conn():
                if settings.run.enable_concurent_sqlite:
                    logger.warning(
                        "You are running parallel execution on top of sqlite database at %s! (see run.enable_concurent_sqlite)",
                        airflow_sql_conn_url(),
                    )
                else:
                    # in theory sqlite can support a decent amount of parallelism, but in practice
                    # the way airflow works each process holds the db exlusively locked which leads
                    # to sqlite DB is locked exceptions
                    raise friendly_error.execute_engine.parallel_or_remote_sqlite(
                        task_executor_type
                    )

        if is_plugin_enabled("dbnd-docker"):
            from dbnd_docker.kubernetes.kubernetes_engine_config import (
                KubernetesEngineConfig,
            )

            if (
                submit_tasks
                and isinstance(remote_engine, KubernetesEngineConfig)
                and run_config.enable_airflow_kubernetes
            ):
                if task_executor_type != AirflowTaskExecutorType.airflow_kubernetes:
                    logger.info("Using dedicated kubernetes executor for this run")
                    task_executor_type = AirflowTaskExecutorType.airflow_kubernetes
                    parallel = True
    else:
        if parallel:
            logger.warning("Airflow is not installed, parallel mode is not supported")

    all_executor_types = [TaskExecutorType.local]
    if is_airflow_enabled():
        from dbnd_airflow.executors import AirflowTaskExecutorType

        all_executor_types.extend(AirflowTaskExecutorType.all())

    if task_executor_type not in all_executor_types:
        raise DatabandConfigError("Unsupported engine type %s" % task_executor_type)

    return task_executor_type, parallel


def get_task_executor(run, task_executor_type, host_engine, target_engine, task_runs):
    # type: (DatabandRun, str, EngineConfig, EngineConfig, List[TaskRun]) -> TaskExecutor

    if task_executor_type == TaskExecutorType.local:
        return LocalTaskExecutor(
            run,
            task_executor_type=task_executor_type,
            host_engine=host_engine,
            target_engine=target_engine,
            task_runs=task_runs,
        )
    else:
        from dbnd_airflow.dbnd_task_executor.dbnd_task_executor_via_airflow import (
            AirflowTaskExecutor,
        )

        return AirflowTaskExecutor(
            run,
            task_executor_type=task_executor_type,
            host_engine=host_engine,
            target_engine=target_engine,
            task_runs=task_runs,
        )



import types

import attr

from dbnd._core.errors import friendly_error
from targets.marshalling import StrLinesMarshaller, StrMarshaller
from targets.marshalling.marshaller import Marshaller
from targets.values import ValueType


@attr.s
class MarshallerCtrl(object):
    target = attr.ib()
    marshaller = attr.ib()  # type: Marshaller
    value_type = attr.ib()  # type: ValueType

    def load(self, **kwargs):
        value = self._load(**kwargs)
        if isinstance(self.marshaller, StrMarshaller):
            value = self.value_type.parse_from_str(value)
        elif isinstance(self.marshaller, StrLinesMarshaller):
            value = [line.rstrip("\n") for line in value]
            value = self.value_type.parse_from_str_lines(value)
        return value

    def _load(self, **kwargs):
        from targets.dir_target import DirTarget
        from targets.file_target import FileTarget
        from targets.multi_target import MultiTarget

        m = self.marshaller
        target = self.target
        if isinstance(target, FileTarget) and target.fs.isdir(target.path):
            target = DirTarget(target.path + "/", target.fs, config=target.config)

        if isinstance(target, MultiTarget):
            if m.support_multi_target_direct_read:
                return m.target_to_value(target, **kwargs)
        elif isinstance(target, DirTarget):
            if m.support_directory_direct_read and m.support_direct_access(target):
                return m.target_to_value(target, **kwargs)
        elif isinstance(target, FileTarget):
            return m.target_to_value(target, **kwargs)

        partitions = target.list_partitions()
        if len(partitions) == 1:
            return m.target_to_value(partitions[0], **kwargs)

        if not self.value_type.support_merge:
            raise friendly_error.marshaller_no_merge(self, target, partitions)
        # Concatenate all data into one DataFrame
        # We don't want list to be stored in memory
        # however, concat does list() on the iterator as one of the first things
        partitions_values = [m.target_to_value(t, **kwargs) for t in partitions]
        return self.value_type.merge_values(*partitions_values)

    def dump(self, value, **kwargs):
        target = self.target
        from targets.multi_target import MultiTarget

        from targets.marshalling import StrMarshaller, StrLinesMarshaller

        if isinstance(self.marshaller, StrMarshaller):
            mode = kwargs.get("mode")
            if (not mode or "b" not in mode) and value is not None:
                # TODO: sometimes we write to strmarshaller binary format.
                value = self.value_type.to_str(value)
        elif isinstance(self.marshaller, StrLinesMarshaller):
            value = self.value_type.to_str_lines(value)
            kwargs["newline"] = kwargs.pop("newline", True)

        m = self.marshaller
        if isinstance(target, MultiTarget):
            raise friendly_error.targets.dump_to_multi_target(self, value)

        from targets.dir_target import DirTarget

        if isinstance(value, types.GeneratorType):
            if not isinstance(target, DirTarget):
                raise friendly_error.targets.dump_generator_to_file(self)
            for value_partition in value:
                m.value_to_target(
                    target=target.partition(), value=value_partition, **kwargs
                )

            target.mark_success()
            return
        selected_target = target
        if isinstance(target, DirTarget):
            # WRITE TO FILE INSIDE DIR.
            if not m.support_directory_direct_write:
                selected_target = target._write_target
        m.value_to_target(target=selected_target, value=value, **kwargs)
        target.mark_success()

    def load_partitioned(self, **kwargs):
        for t in self.target.list_partitions():
            yield self.marshaller.target_to_value(t, **kwargs)



from typing import Any, Optional

from dbnd._core.utils.basics.nothing import NOTHING, is_defined


def first_not_none(*args):
    # type: (*Any) -> Optional[Any]
    """Returns first non None arg."""
    return next((arg for arg in args if arg is not None), None)


def passthrough_kwargs(current_locals):
    current_locals.pop("self", None)
    kwargs = current_locals.pop("kwargs", {})
    kwargs.update(current_locals)
    return kwargs


def _pop_kwarg(kwargs, key, default=NOTHING, error_msg=None):
    if key in kwargs:
        return kwargs.pop(key)
    if is_defined(default):
        return default
    raise TypeError(error_msg or "%s must be specified" % key)



import logging
import typing

import attr

from dbnd._core.constants import TaskRunState, _DbndDataClass


if typing.TYPE_CHECKING:
    from typing import Optional, List
    from datetime import datetime, date
    from uuid import UUID
    from dbnd._core.parameter.parameter_definition import ParameterDefinition


logger = logging.getLogger(__name__)


@attr.s
class TaskDefinitionInfo(_DbndDataClass):
    task_definition_uid = attr.ib()  # type: UUID

    name = attr.ib()  # type: str
    family = attr.ib()  # type: str
    type = attr.ib()  # type: str

    class_version = attr.ib()  # type: str

    task_param_definitions = attr.ib()  # type: List[ParameterDefinition]

    module_source = attr.ib(default=None)  # type: str
    module_source_hash = attr.ib(default=None)  # type: str

    source = attr.ib(default=None)  # type: str
    source_hash = attr.ib(default=None)  # type: str

    def __repr__(self):
        return "TaskDefinitionInfo(%s)" % self.name


@attr.s
class TaskRunParamInfo(_DbndDataClass):
    parameter_name = attr.ib()  # type: str
    value = attr.ib()  # type: str

    value_origin = attr.ib()  # type: str


@attr.s
class TaskRunInfo(_DbndDataClass):
    task_run_uid = attr.ib()  # type: UUID
    task_run_attempt_uid = attr.ib()  # type: UUID

    task_definition_uid = attr.ib()  # type: UUID
    run_uid = attr.ib()  # type: UUID

    execution_date = attr.ib()  # type: datetime
    task_af_id = attr.ib()  # type: str

    task_id = attr.ib()  # type: str
    task_signature = attr.ib()  # type: str

    name = attr.ib()  # type: str

    env = attr.ib()  # type: str

    command_line = attr.ib()  # type: str
    functional_call = attr.ib()  # type: str

    has_downstreams = attr.ib()  # type: bool
    has_upstreams = attr.ib()  # type: bool

    is_reused = attr.ib()  # type: bool
    is_skipped = attr.ib()  # type: bool

    is_dynamic = attr.ib()  # type: bool
    is_system = attr.ib()  # type: bool

    is_root = attr.ib()  # type: bool

    output_signature = attr.ib()  # type: str
    state = attr.ib()  # type: TaskRunState
    target_date = attr.ib()  # type: Optional[date]

    version = attr.ib()  # type: str

    log_local = attr.ib()  # type: str
    log_remote = attr.ib()  # type: str

    task_run_params = attr.ib()  # type: List[TaskRunParamInfo]

    task_signature_source = attr.ib(default=None)  # type: str

    external_links = attr.ib(default=None)  # type: dict

    def __repr__(self):
        return "TaskRunInfo(%s, %s)" % self.name, self.state


@attr.s(hash=True)
class TargetInfo(_DbndDataClass):
    parameter_name = attr.ib()  # type: str
    path = attr.ib(hash=True)  # type: str
    created_date = attr.ib()  # type: Optional[datetime]
    task_run_uid = attr.ib()  # type: Optional[UUID]


@attr.s
class ErrorInfo(_DbndDataClass):
    msg = attr.ib()  # type: str
    help_msg = attr.ib()  # type: str
    exc_type = attr.ib()  # type: str
    databand_error = attr.ib()  # type: bool
    traceback = attr.ib()  # type: str
    nested = attr.ib()  # type: str
    user_code_traceback = attr.ib()  # type: str
    show_exc_info = attr.ib()  # type: bool


@attr.s
class TaskRunEnvInfo(_DbndDataClass):
    uid = attr.ib()  # type: UUID

    cmd_line = attr.ib()  # type: str
    databand_version = attr.ib()  # type: str
    user_code_version = attr.ib()  # type: str
    user_code_committed = attr.ib()  # type: bool
    project_root = attr.ib()  # type: str

    user_data = attr.ib()  # type: str

    user = attr.ib()  # type: str
    machine = attr.ib()  # type: str

    heartbeat = attr.ib()  # type: datetime



from dbnd._core.configuration.dbnd_config import config


def set_tracking_config_overide(use_dbnd_log=None):
    # 1. create proper DatabandContext so we can create other objects
    track_with_cache = config.getboolean("run", "tracking_with_cache")
    config_for_airflow = {
        "run": {
            "skip_completed": track_with_cache,
            "skip_completed_on_run": track_with_cache,
            "validate_task_inputs": track_with_cache,
            "validate_task_outputs": track_with_cache,
        },  # we don't want to "check" as script is task_version="now"
        "task": {
            "task_in_memory_outputs": not track_with_cache
        },  # do not save any outputs
        "core": {"tracker_raise_on_error": False},  # do not fail on tracker errors
    }
    if use_dbnd_log is not None:
        config_for_airflow["log"] = {"disabled": not use_dbnd_log}
    config.set_values(
        config_values=config_for_airflow, override=True, source="dbnd_tracking_config"
    )

    return



from __future__ import print_function

import logging

from dbnd._core.cli.utils import with_fast_dbnd_context
from dbnd._core.constants import AlertSeverity, RunState, TaskRunState
from dbnd._core.errors.base import DatabandApiError
from dbnd._core.utils.basics.text_banner import TextBanner, safe_tabulate
from dbnd._core.utils.cli import options_dependency, requierd_mutually_exclude_options
from dbnd._vendor import click
from dbnd.api.alerts import (
    create_alert,
    delete_alerts,
    get_alerts_filtered,
    list_job_alerts,
)


logger = logging.getLogger(__name__)


class Severity(click.Choice):
    def __init__(self):
        super(Severity, self).__init__(
            choices=AlertSeverity.values(), case_sensitive=False
        )

    def convert(self, value, param, ctx):
        return super(Severity, self).convert(value.upper(), param, ctx)


class RunStateChoice(click.Choice):
    def __init__(self):
        super(RunStateChoice, self).__init__(
            choices=RunState.all_values(), case_sensitive=False
        )

    def convert(self, value, param, ctx):
        return super(RunStateChoice, self).convert(value.upper(), param, ctx)


class TaskRunStateChoice(click.Choice):
    def __init__(self):
        super(TaskRunStateChoice, self).__init__(
            choices=TaskRunState.all_values(), case_sensitive=False,
        )

    def convert(self, value, param, ctx):
        return super(TaskRunStateChoice, self).convert(value.upper(), param, ctx)


supported_alerts = {
    "maximum_duration": dict(
        value_help="Maximum duration of job run in seconds",
        value_type=click.INT,
        with_task=False,
        alert="RunDurationAlert",
        operator=">",
        metric="Run Duration (seconds)",
    ),
    "maximum_retries": dict(
        value_help="Maximum retries of job",
        value_type=click.INT,
        with_task=False,
        alert="MaxRunRetries",
        operator=">",
        metric="Number of Retries Per Run",
    ),
    "ran_last_x_seconds": dict(
        value_help="Duration in seconds of missing success run from job",
        value_type=click.INT,
        with_task=False,
        alert="RanLastXSecondsAlert",
        operator=">",
        metric="Delay Between Subsequent Runs (seconds)",
    ),
    "run_state": dict(
        value_help="State of the run to alert on",
        value_type=RunStateChoice(),
        with_task=False,
        alert="RunStateAlert",
        operator="==",
        metric="Run State",
    ),
    "task_state": dict(
        value_help="State of the task run to alert on",
        value_type=TaskRunStateChoice(),
        with_task=True,
        alert="TaskStateAlert",
        operator="==",
        metric="Task State",
    ),
    "custom_metric": dict(
        value_help="maximum threshold to match the metric",
        value_type=click.STRING,
        with_task=True,
        alert="CustomMetricAlert",
        operator=None,
        metric=None,
    ),
}


def bind_function(name, alert, with_task, value_help, value_type, operator, metric):
    def get_task_option(ctx):
        parent_params = ctx.parent.command.params
        return [op for op in parent_params if op.name == "task"][0]

    @with_fast_dbnd_context
    @click.option(
        "--value", "-v", help=value_help, type=value_type, required=True,
    )
    @click.pass_context
    def inner_function(ctx, value, metric_name=metric, op=operator):
        # xor is less readable
        if not with_task and ctx.obj.get("task"):
            raise click.BadParameter(
                "Task is set but it is unavailable for this alert",
                param=get_task_option(ctx),
                ctx=ctx,
            )
        if with_task and not ctx.obj.get("task"):
            raise click.MissingParameter(
                ctx=ctx,
                param=get_task_option(ctx),
                message="Task name is required for this alert",
            )

        cmd_create_alert(
            manage_ctx=ctx,
            alert=alert,
            operator=op,
            value=value,
            user_metric=metric_name,
        )

    if metric is None:
        inner_function = click.option(
            "--metric-name",
            "-m",
            help="Costume metric name to alert",
            type=click.STRING,
            required=True,
        )(inner_function)

    if operator is None:
        inner_function = click.option(
            "--op",
            "-o",
            help="your operator",
            type=click.Choice(["==", "!=", ">", ">=", "<", "<="]),
            required=True,
        )(inner_function)

    inner_function.__name__ = name

    return inner_function


@click.group()
@click.pass_context
def alerts(ctx):
    """Manage Databand alerts"""
    pass


CREATE_HELP_MSG = """
Create alert for given job\n
\n
EXAMPLES\n
    dbnd alerts create --job my_job --severity HIGH maximum-duration -v 100\n
    dbnd alerts create --job my_job --severity LOW maximum-retries -v 6\n
    dbnd alerts create --job my_job --severity CRITICAL ran-last-x-seconds -v 600\n
    dbnd alerts create --job my_job --severity MEDIUM run-state -v failed\n
    dbnd alerts create --job my_job --severity MEDIUM --task my_job_task task-state -v cancelled\n
    dbnd alerts create --job my_job --severity CRITICAL --task my_job_task custom-metric --metric-name my_metric -v 100 --op "<="\n
"""


@alerts.group(help=CREATE_HELP_MSG)
@click.option("--severity", "-s", help="Alert severity", type=Severity(), required=True)
@click.option("--job", "-j", help="Job name", type=click.STRING, required=False)
@click.option("--job-id", type=click.STRING, required=False)
@click.option("--task", "-t", help="Task name", type=click.STRING, required=False)
@click.option(
    "--update",
    "-u",
    "uid",
    help="Uid of an existing alert to update",
    type=click.STRING,
    required=False,
)
@click.pass_context
def create(ctx, job, job_id, severity, task, uid):
    """Create or update alerts for given job."""
    ctx.obj = {
        "job": job,
        "job_id": job_id,
        "severity": severity,
        "task": task,
        "uid": uid,
    }


# Registering the create commands
for name, values in supported_alerts.items():
    f = bind_function(name, **values)
    create.command(name=name.replace("_", "-"))(f)


@alerts.command("list")
@click.option("--job", "-j", help="Job name", type=click.STRING, required=True)
@with_fast_dbnd_context
def list_jobs(job):
    """Display alert definitions for given job"""
    try:
        alerts_list = list_job_alerts(job_name=job)
    except LookupError as e:
        logger.error(e)
    else:
        print_table("Alerts configured for {job}".format(job=job), alerts_list)


def print_table(header, alerts_list):
    banner = TextBanner(header)
    banner.write(build_alerts_table(alerts_list))
    logger.info(banner.get_banner_str())


def build_alerts_table(alerts_data):
    extract_keys = (
        "uid",
        "job_name",
        "task_name",
        "custom_name",
        "severity",
        "type",
        "operator",
        "value",
        "user_metric",
    )
    headers = (
        "uid",
        "job",
        "task",
        "name",
        "severity",
        "type",
        "op",
        "value",
        "metric",
    )
    table_data = []
    for alert in alerts_data:
        alert_row = [alert.get(key, "") for key in extract_keys]
        table_data.append(alert_row)
    return safe_tabulate(table_data, headers)


def cmd_create_alert(manage_ctx, alert, operator, value, user_metric):
    try:
        alert_def_uid = create_alert(
            job_name=manage_ctx.obj.get("job"),
            job_id=manage_ctx.obj.get("job_id"),
            task_name=manage_ctx.obj.get("task", None),
            uid=manage_ctx.obj.get("uid", None),
            alert_class=alert,
            severity=manage_ctx.obj["severity"],
            operator=operator,
            value=value,
            user_metric=user_metric,
        )
    except DatabandApiError as e:
        logger.error("failed with - {}".format(e.response))
    else:
        logger.info("Created alert %s", alert_def_uid["uid"])


@alerts.command()
@with_fast_dbnd_context
@options_dependency("job", "name")
@requierd_mutually_exclude_options("uid", "wipe", "job")
@click.option("--uid", "-u", help="alert uid", type=click.STRING)
@click.option("--wipe", help="delete all alerts", is_flag=True)
@click.option("--job", "-j", help="job name", type=click.STRING)
@click.option("--name", "-n", help="alert custom name", type=click.STRING)
def delete(uid, wipe, job, name):
    """Delete alerts"""
    if wipe:
        alerts_list = get_alerts_filtered()
    elif uid:
        alerts_list = get_alerts_filtered(alert_uid=uid)
    else:
        alerts_list = get_alerts_filtered(job_name=job, custom_name=name)
    if alerts_list:
        uids = [alert["uid"] for alert in alerts_list]
        delete_alerts(uids)
        print_table("deleted", alerts_list)
    else:
        logger.error("alerts not found to delete")



import base64

from dbnd._core.current import get_databand_context
from dbnd._core.settings.run_info import RunInfoConfig
from dbnd._core.tracking.backends.channels.abstract_channel import TrackingChannel
from dbnd._core.tracking.backends.channels.protobuf_mixin import ProtobufMixin
from dbnd._core.utils.basics.memoized import cached
from dbnd._core.utils.uid_utils import get_uuid
from dbnd.api.proto.generated.tracking_pb2 import (
    Event,
    PostEventsRequest,
    PostEventsResponse,
)


def str_type(obj):
    return "%s.%s" % (obj.__class__.__module__, obj.__class__.__name__)


class TrackingProtoWebChannel(ProtobufMixin, TrackingChannel):
    """Proto API client implementation."""

    @property
    @cached()
    def client(self):
        return get_databand_context().databand_api_client

    @property
    @cached()
    def source_version(self):
        return RunInfoConfig().source_version

    def _handle(self, name, data):
        schema = self.get_schema_by_handler_name(name)

        labels = {
            "sender": "TrackingProtoWebChannel",
            "source_version": self.source_version,
            # TODO: add env details
        }

        event = Event(uuid=str(get_uuid()), schema=str_type(schema), labels=labels)
        event.data.update(data)

        post_event_request = PostEventsRequest()
        post_event_request.events.append(event)
        post_event_request.timestamp.GetCurrentTime()

        data = post_event_request.SerializeToString()

        encoded_response_str = self.client.api_request("tracking/proto", data)
        b64encoded_response_bytes = encoded_response_str.encode("utf-8")
        raw_bytes = base64.b64decode(b64encoded_response_bytes)

        post_event_response = PostEventsResponse()
        post_event_response.ParseFromString(raw_bytes)

        if post_event_response.exception:
            raise Exception("Response error: %s" % post_event_response.exception)

        return post_event_response.responses[event.uuid]

    def is_ready(self):
        return self.client.is_ready()

    def __str__(self):
        return "ProtoWeb"



import pytest

from mock import patch

from dbnd import dbnd_run_stop
from dbnd._core.configuration.environ_config import reset_dbnd_project_config
from dbnd._core.inplace_run.airflow_dag_inplace_tracking import AirflowTaskContext
from dbnd._core.utils.timezone import utcnow


@pytest.fixture
def set_airflow_context():
    with patch(
        "dbnd._core.inplace_run.airflow_dag_inplace_tracking.try_get_airflow_context"
    ) as m:
        try:
            reset_dbnd_project_config()

            m.return_value = AirflowTaskContext(
                dag_id="test_dag",
                task_id="test_task",
                execution_date=utcnow().isoformat(),
            )
            yield
        finally:
            # ensure dbnd_run_stop() is called (normally should happen on exit() )
            dbnd_run_stop()
            reset_dbnd_project_config()



from dbnd._core.errors import friendly_error
from dbnd._core.utils import json_utils
from targets.base_target import Target
from targets.types import Path, PathStr
from targets.values.value_type import ValueType


class _TargetValueType(ValueType):  # extracted for generics support
    """
      ValueType whose value is a ```target```.

      In the task definition, use

      .. code-block:: python

          class MyTask(dbnd.Task):
              some_input = TargetValueType()

              def run(self):
                    self.some_input.open('r')

      """

    support_from_str = False
    config_name = "target"

    def normalize(self, value):
        from dbnd._core.utils.task_utils import to_targets

        return to_targets(value)

    def normalize_to_target(self, value):
        from dbnd._core.utils.task_utils import to_targets

        return to_targets(value)

    def to_str(self, x):
        # value should be the "deferred value" -> Target of any kind
        return json_utils.dumps_canonical(x)

    def save_to_target(self, target, value, **kwargs):
        # if data - automatic mode, we don't give value_type.
        # That makes parameter is agnostic to data type.
        target.dump(value, value_type=None, **kwargs)

    def target_to_value(self, target):
        return target

    def parse_value(self, value, load_value=None, target_config=None):
        value = super(_TargetValueType, self).parse_value(
            value, load_value=load_value, target_config=target_config
        )
        try:
            return self.normalize_to_target(value)
        except Exception as ex:
            raise friendly_error.task_parameters.failed_to_convert_to_target_type(
                self, value, ex
            )

    def load_from_target(self, target, **kwargs):
        return self.target_to_value(target)


class TargetValueType(_TargetValueType):  # _TargetValueType[DataTarget]
    type = Target
    load_on_build = False
    type_str_extras = ("DataTarget", "FileTarget", "DirTarget")

    def is_handler_of_type(self, type_):
        return super(TargetValueType, self).is_handler_of_type(type_) or issubclass(
            type_, Target
        )

    def is_type_of(self, value):
        return isinstance(value, Target)

    def target_to_value(self, target):
        return target


class TargetPathValueType(_TargetValueType):  # _TargetValueType[PathStr]
    type = PathStr
    load_on_build = False

    def target_to_value(self, target):
        return target.path


class TargetPathLibValueType(_TargetValueType):  # _TargetValueType[PathStr]
    type = Path
    load_on_build = False

    def target_to_value(self, target):
        return Path(target.path)



import logging

from dbnd._core.configuration import get_dbnd_project_config
from dbnd._core.constants import TaskType
from dbnd._core.decorator.func_task_call import TaskCallState
from dbnd._core.decorator.schemed_result import FuncResultParameter
from dbnd._core.errors.friendly_error.task_execution import (
    failed_to_assign_result,
    failed_to_process_non_empty_result,
)
from dbnd._core.task.pipeline_task import PipelineTask
from dbnd._core.task.python_task import PythonTask
from dbnd._core.task.task import Task


logger = logging.getLogger(__name__)


class _DecoratedTask(Task):
    _dbnd_decorated_task = True
    result = None

    def _invoke_func(self, extra_kwargs=None, force_invoke=False):
        # this function is in charge of calling user defined code (decorated function) call
        # usually it's called from from task.run/task.band
        extra_kwargs = extra_kwargs or {}
        spec = self._conf__decorator_spec
        invoke_kwargs = {}
        for name in spec.args:
            # if there is no parameter - it was disabled at TaskDefinition building stage
            if self._params.get_param(name) is None:
                continue
            invoke_kwargs[name] = getattr(self, name)
        self.task_user_obj = None

        invoke_kwargs.update(extra_kwargs)
        if spec.is_class:
            obj_cls = self._callable_item
            invoke_kwargs["__force_invoke"] = force_invoke
            # we will get to Metaclass  ( same one that created current `self`)
            # this time  we want it to run user code directly
            self.task_user_obj = obj_cls(**invoke_kwargs)
            try:
                setattr(self.task_user_obj, "_dbnd_task", self)
            except Exception:
                pass
            result = self.task_user_obj.run()
        else:
            # we are going to run user function
            if not self._dbnd_call_state:
                self._dbnd_call_state = TaskCallState()
            self._dbnd_call_state.start()
            func_call = spec.item
            result = func_call(**invoke_kwargs)
            self._dbnd_call_state.finish(result)

        result_param = self.__class__.result
        if result_param is None and result:
            raise failed_to_process_non_empty_result(self, result)

        # spread result into relevant fields.
        if isinstance(result_param, FuncResultParameter):
            # assign all returned values to relevant band Outputs
            if result is None:
                raise failed_to_assign_result(self, result_param)
            for r_name, value in result_param.named_results(result):
                setattr(self, r_name, value)
        else:
            self.result = result
        return result

    def on_kill(self):
        task_user_obj = getattr(self, "task_user_obj", None)
        if task_user_obj is not None and hasattr(task_user_obj, "on_kill"):
            task_user_obj.on_kill()
            return
        else:
            super(_DecoratedTask, self).on_kill()


class DecoratedPythonTask(PythonTask, _DecoratedTask):
    _conf__task_type_name = TaskType.python

    def run(self):
        self._invoke_func(force_invoke=True)


class DecoratedPipelineTask(PipelineTask, _DecoratedTask):
    _conf__task_type_name = TaskType.pipeline

    def band(self):
        return self._invoke_func()



import logging
import typing

from dbnd._core.errors.errors_utils import log_exception
from targets.values.builtins_values import (
    BoolValueType,
    CallableValueType,
    FloatValueType,
    IntValueType,
    NullableStrValueType,
    ObjectValueType,
    StrValueType,
    ValueType,
)
from targets.values.datetime_value import DateTimeValueType, DateValueType
from targets.values.registry import ValueTypeRegistry
from targets.values.structure import (
    DictValueType,
    ListValueType,
    SetValueType,
    TupleValueType,
)
from targets.values.target_config_value import TargetConfigValueType
from targets.values.target_values import (
    TargetPathLibValueType,
    TargetPathValueType,
    TargetValueType,
)
from targets.values.timedelta_value import DateIntervalValueType, TimeDeltaValueType
from targets.values.value_type import InlineValueType
from targets.values.version_value import VersionValueType


if typing.TYPE_CHECKING:
    from typing import Any, Optional
    from targets.value_meta import ValueMetaConf, ValueMeta


# Note: order matters. Examples:
# isinstance(True, int) == True, so it's important to have bool check before int
# isinstance(datetime.datetime.utc(), date) == True
if typing.TYPE_CHECKING:
    from typing import Optional, Any
    from targets.value_meta import ValueMeta, ValueMetaConf

logger = logging.getLogger(__name__)


known_values = []
try:
    import matplotlib
    from targets.values.matplotlib_values import MatplotlibFigureValueType

    known_values.append(MatplotlibFigureValueType())
except ImportError:
    pass

try:
    import pandas
    import numpy
    from targets.values.pandas_values import (
        DataFrameValueType,
        PandasSeriesValueType,
        DataFramesDictValueType,
    )

    known_values.append(DataFrameValueType())
    known_values.append(PandasSeriesValueType())
    known_values.append(DataFramesDictValueType())
except ImportError:
    pass

try:
    import numpy
    from targets.values.numpy_values import NumpyArrayValueType

    known_values.append(NumpyArrayValueType())
except ImportError:
    pass

known_values.extend(
    [
        # simple
        BoolValueType(),
        IntValueType(),
        FloatValueType(),
        # date/time
        DateValueType(),
        DateTimeValueType(),
        TimeDeltaValueType(),
        DateIntervalValueType(),
        # structs,
        ListValueType(),
        DictValueType(),
        SetValueType(),
        TupleValueType(),
        # targets, path
        TargetValueType(),
        TargetPathValueType(),
        TargetPathLibValueType(),
        TargetConfigValueType(),
        CallableValueType(),
        # str types
        VersionValueType(),
        StrValueType(),
        NullableStrValueType(),
    ]
)


# OBJECT VALUE is always the last
known_values.append(ObjectValueType())

_VALUE_TYPES = ValueTypeRegistry(known_value_types=known_values)


def get_types_registry():
    # type: ()-> ValueTypeRegistry
    return _VALUE_TYPES


def get_value_type_of_obj(obj, default_value_type=None):
    # type: (object, Optional[ValueType]) -> Optional[ValueType]
    return get_types_registry().get_value_type_of_obj(obj, default_value_type)


def get_value_type_of_type(obj_type, inline_value_type=False):
    return get_types_registry().get_value_type_of_type(
        obj_type, inline_value_type=inline_value_type
    )


def register_value_type(value_type):
    return get_types_registry().register_value_type(value_type)


def get_value_meta_from_value(
    value_name, value, meta_conf, raise_on_error=False
):  # type: (str, Any, ValueMetaConf, bool) -> Optional[ValueMeta]
    if value is None:
        return None

    obj_value_type = get_value_type_of_obj(value, default_value_type=ObjectValueType())
    if obj_value_type is None:
        logger.info(
            "Can't detect known type for '%s' with type='%s' ", value_name, type(value)
        )
        return None
    try:
        return obj_value_type.get_value_meta(value, meta_conf=meta_conf)

    except Exception as ex:
        log_exception(
            "Failed to get value meta info for '%s' with type='%s'"
            " ( detected as %s)" % (value_name, type(value), obj_value_type),
            ex,
            non_critical=not raise_on_error,
        )
        if raise_on_error:
            raise
    return None



import inspect
import logging
import typing

from collections import OrderedDict
from typing import Any, Dict, Type

import six

from six import iteritems

from dbnd._core.configuration.config_readers import parse_and_build_config_store
from dbnd._core.constants import RESULT_PARAM
from dbnd._core.decorator.func_params_builder import FuncParamsBuilder
from dbnd._core.decorator.schemed_result import FuncResultParameter
from dbnd._core.parameter.parameter_builder import ParameterFactory
from dbnd._core.parameter.parameter_definition import (
    ParameterDefinition,
    _ParameterKind,
)
from dbnd._core.task_build.task_params import TaskDefinitionParams
from dbnd._core.task_build.task_passport import TaskPassport
from dbnd._core.utils.basics.nothing import is_defined
from dbnd._core.utils.structures import combine_mappings
from dbnd._core.utils.uid_utils import get_uuid


if typing.TYPE_CHECKING:
    from dbnd._core.task import Task
    from dbnd._core.task_build.task_metaclass import TaskMetaclass

logger = logging.getLogger(__name__)


def _ordered_params(x):
    return OrderedDict(sorted(x.items(), key=lambda y: y[1].parameter_id))


class TaskDefinition(object):
    """
    This task represents Task Definition

    """

    def __init__(self, task_class, classdict):
        super(TaskDefinition, self).__init__()

        self.task_definition_uid = get_uuid()
        self.hidden = False

        self.task_class = task_class  # type: Type[Task]

        self.task_passport = TaskPassport.from_task_cls(task_class)

        # TODO: maybe use properties or other way to delegate those...
        self.full_task_family = self.task_passport.full_task_family
        self.full_task_family_short = self.task_passport.full_task_family_short
        self.task_family = self.task_passport.task_family
        self.task_config_section = self.task_passport.task_config_section

        # the defaults attribute
        self.defaults = dict()  # type: Dict[ParameterDefinition, Any]

        class_task_params, self.defaults = self._calculate_task_class_values(classdict)

        func_params_builder = self._get_func_params(class_task_params)

        # this is the place we add parameters from function definition
        self._task_params = TaskDefinitionParams(
            class_params=self._update_and_evolve_owner(
                func_params_builder.decorator_kwargs_params, class_task_params
            ),
            user_params=self._update_and_evolve_owner(
                func_params_builder.func_spec_params, {}
            ),
            user_result_params=self._update_and_evolve_owner(
                func_params_builder.result_params, {}
            ),
            is_decorated=self.task_class._conf__decorator_spec is not None,
        )  # type: TaskDefinitionParams

        # if we have output params in function arguments, like   f(some_p=parameter.output)
        # the new function can not return the result of return
        result_param = self._task_params.result_param
        self.single_result_output = (
            result_param is not None
            and self._is_result_single_output(
                self.all_task_params.values(), result_param
            )
        )

        defaults = {
            p.name: p.default
            for p in self.all_task_params.values()
            if is_defined(p.default)
        }
        self.task_defaults_config_store = parse_and_build_config_store(
            source=self.task_passport.format_source_name("defaults"),
            config_values={self.task_config_section: defaults},
            set_if_not_exists_only=True,
        )

        self.task_defaults_config_store.update(
            parse_and_build_config_store(
                source=self.task_passport.format_source_name("defaults_section"),
                config_values=self.defaults,
            )
        )
        # now, if we have overloads in code ( calculated in task_definition):
        # class T(BaseT):
        #     some_base_t_property = new_value
        if self.task_class._conf__track_source_code:
            self.task_source_code = _get_task_source_code(self.task_class)
            self.task_module_code = _get_task_module_source_code(self.task_class)
            self.task_source_file = _get_source_file(self.task_class)
        else:
            self.task_source_code = None
            self.task_module_code = ""
            self.task_source_file = None

    @property
    def all_task_params(self):
        return self._task_params.all_params

    @property
    def class_params(self):
        return self._task_params.class_params

    @property
    def user_params(self):
        return self._task_params.user_params

    def _calculate_task_class_values(self, classdict):
        # reflect inherited attributes
        params, base_defaults = self._discover_base_attributes()

        defaults = combine_mappings(base_defaults, classdict.get("defaults", None))

        self._update_and_evolve_owner(classdict, params)

        return params, defaults

    def _update_and_evolve_owner(self, func_params, params):
        # let update params with new class attributes
        self._update_params_from_attributes(func_params, params)
        # add parameters config
        evolved_params = self._evolve_with_owner(params)
        return evolved_params

    def _get_func_params(self, base_params):
        func_params_builder = FuncParamsBuilder(
            base_params=base_params,
            decorator_spec=self.task_class._conf__decorator_spec,
        )
        func_params_builder.build_func_params()
        return func_params_builder

    def _evolve_with_owner(self, params):
        evolved = {
            name: param.evolve_with_owner(task_cls=self.task_class, name=name)
            for name, param in six.iteritems(params)
        }
        return _ordered_params(evolved)

    def _discover_base_attributes(self):
        # type: ()-> (Dict[str,ParameterDefinition], Dict[str, Any])
        params = dict()
        defaults = dict()
        # we process only "direct" inheritance
        # params will contain definition of param, even it's was overrided by the parent task
        for c in reversed(self.task_class.__bases__):  # type: TaskMetaclass
            if not hasattr(c, "task_definition"):
                logger.debug(
                    "you should inherit from Task objects only: %s -> %s ",
                    self.task_class,
                    c,
                )
                continue
            base_schema = c.task_definition  # type: TaskDefinition
            defaults = combine_mappings(defaults, base_schema.defaults)
            params = combine_mappings(params, c.task_definition.all_task_params)

        return params, defaults

    def _update_params_from_attributes(self, classdict, params):
        class_values = dict()

        for a_name, a_obj in iteritems(classdict):
            context = "%s.%s" % (self.task_family, a_name)
            try:
                if isinstance(a_obj, ParameterFactory):
                    params[a_name] = a_obj.build_parameter(context)
                elif isinstance(a_obj, ParameterDefinition):
                    params[a_name] = a_obj
                else:
                    class_values[a_name] = a_obj
            except Exception:
                logger.error("Failed to process %s" % context)
                raise

        for p_name, p_val in iteritems(class_values):
            if p_name not in params:
                continue
            params[p_name] = params[p_name].modify(default=p_val)

    def _is_result_single_output(self, params, result):
        """
        check that task has only one output and it's output is result
         (there can be sub outputs that are part of result)
        """
        if not result:
            return False
        names = result.names if isinstance(result, FuncResultParameter) else []
        for p in params:
            if p.system or p.kind != _ParameterKind.task_output:
                continue
            if p.name in [RESULT_PARAM, "task_band"]:
                continue
            if p.name in names:
                continue
            return False
        return True

    def __str__(self):
        return "TaskDefinition(%s)" % (self.full_task_family)


def _get_task_source_code(task):
    try:
        import inspect

        if hasattr(task, "_conf__decorator_spec") and task._conf__decorator_spec:
            item = task._conf__decorator_spec.item
            return inspect.getsource(item)
        else:
            return inspect.getsource(task)
    except (TypeError, OSError):
        logger.debug("Failed to task source for %s", task)
    except Exception:
        logger.debug("Error while getting task source")
    return "Error while getting task source"


def _get_task_module_source_code(task):
    try:
        return inspect.getsource(inspect.getmodule(task))
    except TypeError:
        logger.debug("Failed to module source for %s", task)
    except Exception:
        logger.exception("Error while getting task module source")
    return "Error while getting task source"


def _get_source_file(task):
    try:
        return inspect.getfile(task.__class__).replace(".pyc", ".py")
    except Exception:
        logger.warning("Failed find a path of source code for task {}".format(task))
    return None



from dbnd._core.constants import TaskType
from dbnd._core.task.data_task import Task


class PythonTask(Task):
    """
    Executes a Python code

    """

    _conf__task_type_name = TaskType.python



import typing

from dbnd._vendor import pluggy


if typing.TYPE_CHECKING:
    from dbnd._core.task_run.task_run import TaskRun


hookspec = pluggy.HookspecMarker("dbnd")


@hookspec
def dbnd_setup_plugin():
    """ Called right after plugin is imported  """
    pass


@hookspec
def dbnd_get_commands():
    """ Called from main cli entry point """
    pass


@hookspec
def dbnd_on_pre_init_context(ctx):
    """ Called from DatabandContext before entering a new context """
    pass


@hookspec
def dbnd_on_new_context(ctx):
    """ Called from DatabandContext when entering a new context """
    pass


@hookspec
def dbnd_on_existing_context(ctx):
    """ Called from DatabandContext when entering an existing context """
    pass


@hookspec
def dbnd_post_enter_context(ctx):
    """ Called from DatabandContext after enter initialization steps """
    pass


@hookspec
def dbnd_on_exit_context(ctx):
    """ Called from DatabandContext when exiting """
    pass


@hookspec
def dbnd_setup_unittest():
    """ Called when running in test mode  """
    pass


@hookspec
def dbnd_task_run_context(task_run):
    # type: (TaskRun)-> Context
    """ Using this context when running task_run  """
    pass


@hookspec
def dbnd_build_project_docker(docker_engine, docker_build_task):
    """ Runs on docker build """
    pass



from dbnd._core.errors import DatabandConfigError
from dbnd._core.utils.http.constants import AUTHS_SUPPORTED, NO_AUTH


class Endpoint(object):
    def __init__(
        self, url, auth=NO_AUTH, username="", password="", implicitly_added=False
    ):
        if not url:
            raise DatabandConfigError(u"URL must not be empty")
        if auth not in AUTHS_SUPPORTED:
            raise DatabandConfigError(u"Auth '{}' not supported".format(auth))

        self.url = url.rstrip(u"/")
        self.username = username
        self.password = password
        self.auth = auth
        # implicitly_added is set to True only if the endpoint wasn't configured manually by the user through
        # a widget, but was instead implicitly defined as an endpoint to a wrapper kernel in the configuration
        # JSON file.
        self.implicitly_added = implicitly_added

    def __eq__(self, other):
        if type(other) is not Endpoint:
            return False
        return (
            self.url == other.url
            and self.username == other.username
            and self.password == other.password
            and self.auth == other.auth
        )

    def __hash__(self):
        return hash((self.url, self.username, self.password, self.auth))

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return u"Endpoint({})".format(self.url)



"""
Task build process
=================

When a task is created:
 * We calculate task_env and task_config first
 * for every property, we look at all possible sections ( full task name, task name, and inherited tasks)
 * Config Layers:
     system layer ( starting point, contains Configuration, Environment, and Command-line
     new layer from task.task_definition.defaults (Task.defaults)  merge_strategy=on_non_exist_only
     updated with overrides from constructor
     updated with task_config
     updated with task_env values

Special properties of the Task
    Task.defaults is the lowest level of config (DEFAULTS)
     * definition: custom property of the Task object
     * definition: all values are merged when TaskB inherits from TaskA (at TaskDefinition)
     * task.defaults value is calculated at TaskDefinition during the "compilation of the Task class definition"
     * usage: its value is merged into Current config (only updates nonexisting fields) at TaskMetaClass
     * do not support override()!!!
     * there is no "merge" behavior for values. (Dictionaries are going to be replaced)
         class TaskA(...):
            defaults = {SparkConf.conf : { 1:1} , SparkConf.jar : "taskajar }

         class TaskB(TaskA):
            defaults = {SparkConf.conf : { 1:2}  }
        TaskB.defaults == {SparkConf.conf : { 1:2} , SparkConf.jar : "taskajar }

    Task.task_config - overrides current config level
     * definition: regular parameter of the task
     * definition: it "replaces" inherited task_config from super class if exists
     * its value is calculated every time we create Task at TaskMeta
     * usage: it updates current config layer (
     * supports override():  `task_config = {SparkConf.jar: override("my_jar")}`
     * Usually, it's used with some basic "nested" config param: like SparkConfig.conf.
     Task.task_config issues:
       1.     task_config = {SparkConf.jar: "my_jar"}
             [spark_remote]
             jar=from_config
        User will get my_jar when engine is [spark] and  "from_config" if engine is [spark_remote].
        The reason is that engine looks for the value in most relevant section which is [spark_remote],
        while SparkConf.jar provides a value for [spark] section
        The only way to workaround it right now is to use:
            task_config = {SparkConf.jar: override("my_jar")}
        The system will look for override values and if it exists,
        the override will be taken (despite best section match).

    Task(overrides=conf_value...)
     * uses the same system as task_config, but override is applied to all values

"""



from typing import Dict, Generic, Optional, TypeVar

import attr

from dbnd._core.constants import RESULT_PARAM
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.parameter.parameter_value import ParameterValue


def _ensure_dict(val):
    if isinstance(val, dict):
        return val
    if isinstance(val, list):
        return {p.name: p for p in val}
    if not val:
        return {}
    raise NotImplementedError()


def _merge_dicts(*dicts):
    merged = {}
    for d in dicts:
        merged.update(d)
    return merged


T = TypeVar("T")


@attr.s
class _TaskParams(Generic[T]):
    _class_params = attr.ib(factory=dict, converter=_ensure_dict)
    _user_params = attr.ib(factory=dict, converter=_ensure_dict)
    _user_result_params = attr.ib(factory=dict, converter=_ensure_dict)

    is_decorated = attr.ib(factory=bool)

    @property
    def result_param(self):
        # type: () -> Optional[T]
        if self.is_decorated:
            return self._user_result_params.get(RESULT_PARAM)
        return self._class_params.get(RESULT_PARAM)

    @property
    def all_params(self):
        # type: () -> Dict[str, T]
        return _merge_dicts(
            self._class_params, self._user_params, self._user_result_params
        )

    @property
    def class_params(self):
        # type: () -> Dict[str, T]
        return _merge_dicts(self._class_params, self._user_result_params)

    @property
    def user_params(self):
        # type: () -> Dict[str, T]
        return self._user_params

    def get_any_param(self, name, default=None):
        # will get param by name regardless of its type
        for g in (self._user_params, self._user_result_params, self._class_params):
            if name in g:
                return g[name]
        return default

    def clone(self):
        # type: () -> _TaskParams[T]
        return self.__class__(
            class_params=dict(self._class_params),
            user_params=dict(self._user_params),
            user_result_params=dict(self._user_result_params),
        )


TaskDefinitionParams = _TaskParams[ParameterDefinition]
TaskValueParams = _TaskParams[ParameterValue]



from six.moves.builtins import str


def parse_bool(s):
    if s is None:
        return False

    val = str(s).lower().strip()
    if "#" in val:
        val = val.split("#")[0].strip()
    if val.lower() in ("t", "true", "yes", "1"):
        return True
    elif val.lower() in ("f", "false", "no", "0"):
        return False
    else:
        raise ValueError("Can't parse '%s' as boolean" % s)


def indent(string, prefix):
    """ a backward compatible textwrap.indent replacement """
    if not string:
        return ""

    return "\n".join(prefix + l for l in string.splitlines())



import logging

from dbnd._core.current import get_databand_context, is_verbose
from dbnd._core.utils.basics.nothing import NOTHING, is_defined
from dbnd._core.utils.git import get_git_commit
from dbnd._core.utils.project.project_fs import project_path
from dbnd._core.utils.timezone import utcnow
from targets.values import StrValueType


logger = logging.getLogger(__name__)


class VersionAlias(object):
    now = "now"
    context_uid = "context_uid"
    git = "git"


class VersionStr(str):
    pass


class VersionValueType(StrValueType):
    """
    A Value whose value is a regular string,
    supports aliases from VersionAlias
    """

    type = VersionStr
    discoverable = False

    def parse_from_str(self, s):
        s = super(VersionValueType, self).parse_from_str(s)
        if not s:
            return s

        if s.lower() == VersionAlias.now:
            return utcnow().strftime("%Y%m%d_%H%M%S")

        if s.lower() == VersionAlias.context_uid:
            return get_databand_context().current_context_uid

        if s.lower() == VersionAlias.git:
            return str(get_project_git())

        return s


_project_git_version = NOTHING


def get_project_git():
    global _project_git_version
    if is_defined(_project_git_version):
        return _project_git_version

    _project_git_version = get_git_commit(project_path(), verbose=is_verbose())
    return _project_git_version



from collections import Mapping

import six

from dbnd._core.utils.type_check_utils import is_instance_by_class_name


class DatabandDict(object):
    pass


def flatten(struct):
    """
    Creates a flat list of all all items in structured output (dicts, lists, items):

    .. code-block:: python

        >>> sorted(flatten({'a': 'foo', 'b': 'bar'}))
        ['bar', 'foo']
        >>> sorted(flatten(['foo', ['bar', 'troll']]))
        ['bar', 'foo', 'troll']
        >>> flatten('foo')
        ['foo']
        >>> flatten(42)
        [42]
    """
    if struct is None:
        return []
    flat = []
    if isinstance(struct, (dict, Mapping)):
        for _, result in six.iteritems(struct):
            flat += flatten(result)
        return flat
    if isinstance(struct, six.string_types):
        return [struct]

    if getattr(struct, "target_no_traverse", None):
        return [struct]

    try:
        # if iterable
        iterator = iter(struct)
    except TypeError:
        return [struct]

    for result in iterator:
        flat += flatten(result)
    return flat


def f_identity(x):
    return x


def f_not_none(x):
    return x


def f_no_filter(x):
    return True


def traverse(
    struct,
    convert_f=f_identity,
    filter_none=False,
    filter_empty=False,
    convert_types=None,
):
    """
    Maps all Tasks in a structured data object to their .output().
    :param convert_types non basic types to apply convert_f
    :param filter_none  remove None from structures
    """

    def t(obj):
        if convert_types and isinstance(obj, convert_types):
            return convert_f(obj)

        if isinstance(obj, Mapping):
            # noinspection PyArgumentList
            converted = ((k, t(v)) for k, v in six.iteritems(obj))
            if filter_none:
                converted = ((k, v) for k, v in converted if v is not None)
            new_obj = obj.__class__(converted)
            if filter_empty and not new_obj:
                return None
            return new_obj

        if isinstance(obj, six.string_types):
            return convert_f(obj)

        if "pyspark" in str(type(obj)):
            return convert_f(obj)

        try:
            target_no_traverse = hasattr(obj, "target_no_traverse") and getattr(
                obj, "target_no_traverse", None
            )
        except ValueError:
            # SPARK OBJECTS do not support hasattr
            target_no_traverse = True
        if target_no_traverse is bool and target_no_traverse:
            return convert_f(obj)

        list_obj_constructor = None
        if isinstance(obj, (list, tuple, set)):
            list_obj_constructor = obj.__class__
        elif is_instance_by_class_name(obj, "DataFrame"):
            pass
        else:
            try:
                iter(obj)  # noqa: F841
                list_obj_constructor = list
            except TypeError:
                pass

        # we can parse and reconstruct list object
        if list_obj_constructor:
            converted = (t(r) for r in obj)
            if filter_none:
                converted = filter(lambda x: x is not None, converted)
            new_obj = list_obj_constructor(converted)
            if filter_empty and not new_obj:
                return None
            return new_obj

        # so it's simple obj, let apply function
        return convert_f(obj)

    return t(struct)


def _frozen_set(struct):
    if isinstance(struct, set):
        from dbnd._core.utils import json_utils

        return sorted(struct, key=lambda x: json_utils.dumps_canonical(x))
    return struct


def getpaths(struct):
    """
    Maps all Tasks in a structured data object to their .output().
    """
    return traverse(struct, lambda t: t.task_outputs)


def traverse_to_str(obj):
    if obj is None:
        return None
    return traverse(obj, convert_f=str)


def traverse_frozen_set(obj):
    return traverse(obj, convert_types=(set,), convert_f=_frozen_set)



import os


def abs_join(*path):
    return os.path.abspath(os.path.normpath(os.path.join(*path)))


def relative_path(file_in_path, *path):
    current_dir = os.path.dirname(os.path.abspath(file_in_path))
    return abs_join(current_dir, *path)


def relative_path_directory(file_in_path, *path):
    current_dir = os.path.abspath(file_in_path)
    return abs_join(current_dir, *path)



import io
import os

from targets.config import get_local_tempfile


class AtomicLocalFile(io.BufferedWriter):
    """Class to create a Target that creates
    a temporary file in the local filesystem before
    moving it to its final destination.

    This class is just for the writing part of the Target. See
    :class:`targets.file.FileTarget` for example
    """

    def __init__(self, path, fs, mode="w", overwrite=False):
        self.__tmp_path = self.generate_tmp_path(path)
        self.path = path
        self.fs = fs
        file_io = io.FileIO(self.__tmp_path, mode)
        self.overwrite = overwrite

        super(AtomicLocalFile, self).__init__(file_io)

    def close(self):
        super(AtomicLocalFile, self).close()
        self.move_to_final_destination()

    def generate_tmp_path(self, path):
        return get_local_tempfile(os.path.basename(path))

    def move_to_final_destination(self):
        self.fs.move_from_local(self.tmp_path, self.path, overwrite=self.overwrite)

    def __del__(self):
        # this is required on Windows, otherwise os.remove fails
        if not self.closed:
            super(AtomicLocalFile, self).close()
        if os.path.exists(self.tmp_path):
            os.remove(self.tmp_path)

    @property
    def tmp_path(self):
        return self.__tmp_path

    def __exit__(self, exc_type, exc, traceback):
        " Close/commit the file if there are no exception "
        if exc_type:
            return
        return super(AtomicLocalFile, self).__exit__(exc_type, exc, traceback)


class _DeleteOnCloseFile(io.FileIO):
    def close(self):
        super(_DeleteOnCloseFile, self).close()
        try:
            os.remove(self.name)
        except OSError:
            # Catch a potential threading race condition and also allow this
            # method to be called multiple times.
            pass

    def readable(self):
        return True

    def writable(self):
        return False

    def seekable(self):
        return True



"""
Context: airflow operator/task is running a function with @task
Here we create dbnd objects to represent them and send to webserver through tracking api.
"""
import datetime
import logging
import os
import sys

from collections import Mapping
from typing import Any, Dict, Iterable, Optional, Tuple

import attr

import dbnd._core.utils.basics.environ_utils

from dbnd._core.configuration import environ_config
from dbnd._core.configuration.environ_config import (
    ENV_DBND_TRACKING_ATTEMPT_UID,
    _debug_init_print,
    spark_tracking_enabled,
)
from dbnd._core.constants import UpdateSource
from dbnd._core.current import get_settings
from dbnd._core.errors.errors_utils import UserCodeDetector
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task.task import Task
from dbnd._core.task_build.dynamic import build_dynamic_task_class
from dbnd._core.utils.airflow_cmd_utils import generate_airflow_cmd
from dbnd._core.utils.seven import import_errors
from dbnd._core.utils.type_check_utils import is_instance_by_class_name
from dbnd._core.utils.uid_utils import get_airflow_instance_uid


logger = logging.getLogger(__name__)
_SPARK_ENV_FLAG = "SPARK_ENV_LOADED"  # if set, we are in spark

DAG_SPECIAL_TASK_ID = "DAG"


def override_airflow_log_system_for_tracking():
    return dbnd._core.utils.basics.environ_utils.environ_enabled(
        environ_config.ENV_DBND__OVERRIDE_AIRFLOW_LOG_SYSTEM_FOR_TRACKING
    )


def disable_airflow_subdag_tracking():
    return dbnd._core.utils.basics.environ_utils.environ_enabled(
        environ_config.ENV_DBND__DISABLE_AIRFLOW_SUBDAG_TRACKING, False
    )


@attr.s
class AirflowTaskContext(object):
    dag_id = attr.ib()  # type: str
    execution_date = attr.ib()  # type: str
    task_id = attr.ib()  # type: str
    try_number = attr.ib(default=1)
    context = attr.ib(default=None)  # type: Optional[dict]
    airflow_instance_uid = attr.ib(default=None)  # type: Optional[UUID]

    root_dag_id = attr.ib(init=False, repr=False)  # type: str
    is_subdag = attr.ib(init=False, repr=False)  # type: bool

    def __attrs_post_init__(self):
        self.is_subdag = "." in self.dag_id and not disable_airflow_subdag_tracking()
        if self.is_subdag:
            dag_breadcrumb = self.dag_id.split(".", 1)
            self.root_dag_id = dag_breadcrumb[0]
            self.parent_dags = []
            # for subdag "root_dag.subdag1.subdag2.subdag3" it should return
            # root_dag, root_dag.subdag1, root_dag.subdag1.subdag2
            # WITHOUT the dag_id itself!
            cur_name = []
            for name_part in dag_breadcrumb[:-1]:
                cur_name.append(name_part)
                self.parent_dags.append(".".join(cur_name))
        else:
            self.root_dag_id = self.dag_id
            self.parent_dags = []


def extract_airflow_context(airflow_context):
    # type: (Dict[str, Any]) -> Optional[AirflowTaskContext]
    """Create AirflowTaskContext for airflow_context dict"""

    task_instance = airflow_context.get("task_instance")
    if task_instance is None:
        return None

    dag_id = task_instance.dag_id
    task_id = task_instance.task_id
    execution_date = str(task_instance.execution_date)
    try_number = task_instance.try_number

    if dag_id and task_id and execution_date:
        return AirflowTaskContext(
            dag_id=dag_id,
            execution_date=execution_date,
            task_id=task_id,
            try_number=try_number,
            context=airflow_context,
            airflow_instance_uid=get_airflow_instance_uid(),
        )

    logger.debug(
        "airflow context from inspect, at least one of those params is missing"
        "dag_id: {}, execution_date: {}, task_id: {}".format(
            dag_id, execution_date, task_id
        )
    )
    return None


def try_get_airflow_context():
    # type: ()-> Optional[AirflowTaskContext]
    # first try to get from spark, then from call stack and then from airflow env
    try:
        for func in [
            try_get_airflow_context_from_spark_conf,
            try_get_airflow_context_env,
        ]:
            context = func()
            if context:
                return context
            else:
                msg = func.__name__.replace("_", " ").replace("try", "couldn't")
                logger.debug(msg)

    except Exception:
        return None


def try_get_airflow_context_env():
    # type: ()-> Optional[AirflowTaskContext]
    # Those env vars are set by airflow before running the operator
    dag_id = os.environ.get("AIRFLOW_CTX_DAG_ID")
    execution_date = os.environ.get("AIRFLOW_CTX_EXECUTION_DATE")
    task_id = os.environ.get("AIRFLOW_CTX_TASK_ID")
    try_number = os.environ.get("AIRFLOW_CTX_TRY_NUMBER")
    airflow_instance_uid = os.environ.get("AIRFLOW_CTX_UID")

    if dag_id and task_id and execution_date:
        return AirflowTaskContext(
            dag_id=dag_id,
            execution_date=execution_date,
            task_id=task_id,
            try_number=int(try_number) if try_number else None,
            airflow_instance_uid=airflow_instance_uid,
        )

    logger.debug(
        "airflow context from env, at least one of those environment var is missing"
        "dag_id: {}, execution_date: {}, task_id: {}".format(
            dag_id, execution_date, task_id
        )
    )
    return None


_IS_SPARK_INSTALLED = None


def _is_dbnd_spark_installed():
    global _IS_SPARK_INSTALLED
    if _IS_SPARK_INSTALLED is not None:
        return _IS_SPARK_INSTALLED
    try:
        try:
            from pyspark import SparkContext

            from dbnd_spark import dbnd_spark_bootstrap

            dbnd_spark_bootstrap()
        except import_errors:
            _IS_SPARK_INSTALLED = False
        else:
            _IS_SPARK_INSTALLED = True
    except Exception:
        # safeguard, on any exception
        _IS_SPARK_INSTALLED = False

    return _IS_SPARK_INSTALLED


def try_get_airflow_context_from_spark_conf():
    # type: ()-> Optional[AirflowTaskContext]
    if not spark_tracking_enabled() or _SPARK_ENV_FLAG not in os.environ:
        _debug_init_print(
            "DBND__ENABLE__SPARK_CONTEXT_ENV or SPARK_ENV_LOADED are not set"
        )
        return None

    if not _is_dbnd_spark_installed():
        _debug_init_print("failed to import pyspark or dbnd-spark")
        return None
    try:
        _debug_init_print("creating spark context to get spark conf")
        from pyspark import SparkContext

        conf = SparkContext.getOrCreate().getConf()

        dag_id = conf.get("spark.env.AIRFLOW_CTX_DAG_ID")
        execution_date = conf.get("spark.env.AIRFLOW_CTX_EXECUTION_DATE")
        task_id = conf.get("spark.env.AIRFLOW_CTX_TASK_ID")
        try_number = conf.get("spark.env.AIRFLOW_CTX_TRY_NUMBER")
        airflow_instance_uid = conf.get("spark.env.AIRFLOW_CTX_UID")

        if dag_id and task_id and execution_date:
            return AirflowTaskContext(
                dag_id=dag_id,
                execution_date=execution_date,
                task_id=task_id,
                try_number=try_number,
                airflow_instance_uid=airflow_instance_uid,
            )
    except Exception as ex:
        logger.info("Failed to get airflow context info from spark job: %s", ex)

    return None


class AirflowOperatorRuntimeTask(Task):
    task_is_system = False
    _conf__track_source_code = False

    dag_id = parameter[str]
    execution_date = parameter[datetime.datetime]

    def _initialize(self):
        super(AirflowOperatorRuntimeTask, self)._initialize()
        self.task_meta.task_functional_call = ""

        self.task_meta.task_command_line = generate_airflow_cmd(
            dag_id=self.dag_id,
            task_id=self.task_id,
            execution_date=self.execution_date,
            is_root_task=False,
        )


def build_run_time_airflow_task(af_context):
    # type: (AirflowTaskContext) -> (AirflowOperatorRuntimeTask, str, UpdateSource)
    if af_context.context:
        # we are in the execute entry point and therefore that task name is <task>__execute
        task_family = "%s__execute" % af_context.task_id

        airflow_operator = af_context.context["task_instance"].task

        tracked_function = None
        if is_instance_by_class_name(airflow_operator, "PythonOperator"):
            tracked_function = airflow_operator.python_callable

        # find the template fields of the operators
        user_params = get_flatten_operator_params(airflow_operator)
        if tracked_function:
            user_params["function_name"] = tracked_function.__name__

        # create params definitions for the operator's fields
        params_def = {
            name: parameter[type(value)].build_parameter("inline")
            for name, value in user_params.items()
        }

        task_class = build_dynamic_task_class(
            AirflowOperatorRuntimeTask, task_family, params_def,
        )

        if tracked_function:
            import inspect

            task_class.task_definition.task_source_code = inspect.getsource(
                tracked_function
            )
            task_class.task_definition.task_module_code = inspect.getsource(
                inspect.getmodule(tracked_function)
            )

    else:
        # if this is an inline run-time task, we name it after the script which ran it
        task_family = sys.argv[0].split(os.path.sep)[-1]
        task_class = build_dynamic_task_class(AirflowOperatorRuntimeTask, task_family)
        module_code = get_user_module_code()
        task_class.task_definition.task_module_code = module_code
        task_class.task_definition.task_source_code = module_code
        user_params = {}

    root_task = task_class(
        task_name=task_family,
        dag_id=af_context.dag_id,
        execution_date=af_context.execution_date,
        task_version="%s:%s" % (af_context.task_id, af_context.execution_date),
        **user_params
    )

    job_name = "{}.{}".format(af_context.dag_id, af_context.task_id)
    source = UpdateSource.airflow_tracking
    return root_task, job_name, source


def get_user_module_code():
    try:
        user_frame = UserCodeDetector.build_code_detector().find_user_side_frame(
            user_side_only=True
        )
        if user_frame:
            module_code = open(user_frame.filename).read()
            return module_code
    except Exception as ex:
        return


def should_flatten(operator, attr_name):
    flatten_config = get_settings().tracking.flatten_operator_fields
    for op_name in flatten_config:
        if is_instance_by_class_name(operator, op_name):
            return attr_name in flatten_config[op_name]
    return False


def flatten_param(attr_name, value):
    # type: (str, Any) -> Iterable[Tuple[str, Any]]
    """
    Flatten means we take out the first layer of nested value and split it to elements with relevant name.
    examples:
    >>> flatten_param("name", [0,1, [1,2,3,4]]) == [("name.0", 0),  ("name.1", 1), ("name.2", [1,2,3,4]),]
    >>> flatten_param("other", {"a": 1, "b": {"inner": "value"}}) == [("other.a", 1),  ("other.b", {"inner": "value"})]
    >>> flatten_param("other_name", "1") == [("other_name", "1")]
    """
    if isinstance(value, (dict, Mapping)):
        for name, value in value.items():
            yield ("%s.%s" % (attr_name, name), value)

    elif isinstance(value, list):
        for i, value in enumerate(value):
            yield ("%s.%s" % (attr_name, i), value)

    else:
        yield (attr_name, value)


def get_flatten_operator_params(operator):
    params = []
    for attr_name in operator.template_fields:
        value = getattr(operator, attr_name)
        if should_flatten(operator, attr_name):
            params.extend(flatten_param(attr_name, value))

        else:
            params.append((attr_name, value))

    # we want to make sure that the value is str
    return {name: repr(value) for name, value in params}


def calc_task_run_attempt_key_from_af_ti(ti):
    """
    Creates a key from airflow TaskInstance, for communicating task_run_attempt_uid
    """
    return ":".join(
        [ENV_DBND_TRACKING_ATTEMPT_UID, ti.dag_id, "%s__execute" % ti.task_id]
    )


def calc_task_run_attempt_key_from_dbnd_task(task):
    """
    Creates a key from dbnd Task, for communicating task_run_attempt_uid
    """
    return ":".join(
        [ENV_DBND_TRACKING_ATTEMPT_UID, task.dag_id, task.get_task_family()]
    )


def try_pop_attempt_id_from_env(task):
    """
    if the task is an airflow execute task we try to pop the attempt id from environ
    """
    if isinstance(task, AirflowOperatorRuntimeTask):
        key = calc_task_run_attempt_key_from_dbnd_task(task)
        attempt_id = os.environ.get(key, None)
        if attempt_id:
            del os.environ[key]

        return attempt_id



import logging
import typing

import six

from dbnd import get_dbnd_project_config
from dbnd._core.constants import SystemTaskName
from dbnd._core.current import is_verbose
from dbnd._core.errors import get_help_msg, show_exc_info
from dbnd._core.errors.errors_utils import log_exception, nested_exceptions_str
from dbnd._core.settings import DescribeConfig
from dbnd._core.task.task import Task
from dbnd._core.task_ctrl.task_ctrl import TaskSubCtrl
from dbnd._core.task_run.task_run_error import task_call_source_to_str
from dbnd._core.utils.basics.text_banner import TextBanner, safe_string, safe_tabulate
from dbnd._vendor.termcolor import colored
from targets import DataTarget, InMemoryTarget, Target


if typing.TYPE_CHECKING:
    from dbnd._core.task_run.task_run import TaskRun

logger = logging.getLogger(__name__)


def operators_list_str(operators, separator="\n\t"):
    operators = [o.task_id for o in operators]
    return separator.join(operators)


_TASK_FIELDS = [
    "task_env",
    "task_target_date",
    "task_version",
    "task_class_version",
    "task_enabled",
    "task_enabled_in_prod",
    "class_version",  # deprecated,
    "task_env",
    "task_band",
]


class FormatterVerbosity(object):
    LOW = 0
    NORMAL = 10
    HIGH = 20


def _important_value(value):
    return colored(value, color="red")


def _f_env(env):
    return _important_value(env) if env == "prod" else env


def _f_none_default(value, default):
    if value != default:
        return _important_value(value)
    return value


class TaskVisualiser(TaskSubCtrl):
    def __init__(self, task):
        super(TaskVisualiser, self).__init__(task)

    def banner(
        self,
        msg,
        color=None,
        verbose=False,
        print_task_band=False,
        task_run=None,
        exc_info=None,
    ):
        try:
            banner = TextBanner(msg, color)

            if verbose or is_verbose():
                verbosity = FormatterVerbosity.HIGH
            else:
                verbosity = FormatterVerbosity.NORMAL

            builder = _TaskBannerBuilder(
                task=self.task,
                banner=banner,
                verbosity=verbosity,
                print_task_band=print_task_band,
            )

            # different banners for tracking and orchestration
            if get_dbnd_project_config().is_tracking_mode():
                builder.build_tracking_banner(task_run=task_run, exc_info=exc_info)
            else:
                builder.build_orchestration_banner(task_run=task_run, exc_info=exc_info)

            return banner.get_banner_str()

        except Exception as ex:
            log_exception(
                "Failed to calculate banner for '%s'" % self.task_id,
                ex,
                non_critical=True,
            )
            return msg + (" ( task_id=%s)" % self.task_id)


class _TaskBannerBuilder(TaskSubCtrl):
    """
    Describe how to build a banner for different scenarios.
    Also describe how to build each of the columns in the banner.
    todo: can be extracted to builder and director
    """

    def __init__(self, task, banner, verbosity, print_task_band, task_run=None):
        # type: (_TaskBannerBuilder, Task, TextBanner, int, bool,  ...) -> None
        super(_TaskBannerBuilder, self).__init__(task)
        self.banner = banner
        self.task_run = task_run

        self.verbosity = verbosity
        self.print_task_band = print_task_band

        self.is_driver_or_submitter = (
            self.task.task_name in SystemTaskName.driver_and_submitter
        )

        self.table_director = _ParamTableDirector(task, banner)

    def build_orchestration_banner(self, task_run=None, exc_info=None):
        # type: (_TaskBannerBuilder, TaskRun, ...) -> TextBanner
        """building the banner for orchestration scenario"""

        self.add_task_info()

        if not self.task.ctrl.should_run():
            self.add_disabled_info()

        if task_run:
            self.add_time_info(task_run)
            self.add_tracker_info(task_run)
            self.add_task_run_info(task_run)

            if task_run.external_resource_urls:
                self.add_external_resource_info(task_run)

            self.add_log_info(task_run)

        all_info = self.verbosity >= FormatterVerbosity.HIGH
        self.table_director.add_params_table(
            all_params=all_info,
            param_format=True,
            param_source=True,
            param_section=all_info,
            param_default=all_info,
        )

        if self.verbosity >= FormatterVerbosity.HIGH:
            self._add_verbose_info()

        elif self.print_task_band:
            self._add_task_band_info()

        self.task._task_banner(self.banner, verbosity=self.verbosity)
        self.add_stack_and_errors(exc_info, task_run, self.verbosity)

        return self.banner

    def build_tracking_banner(self, task_run=None, exc_info=None):
        # type: (_TaskBannerBuilder, TaskRun, ...) -> TextBanner
        """building the banner for tracking scenario - keep light as possible"""

        if task_run:
            self.add_tracker_info(task_run)
            self.add_log_info(task_run)

        self.table_director.add_params_table()

        self.add_stack_and_errors(exc_info, task_run, self.verbosity)

        return self.banner

    def add_stack_and_errors(self, exc_info, task_run, verbosity):
        if task_run and task_run.last_error:
            self._task_create_stack()
            self._add_last_error_info(exc_info=task_run.last_error.exc_info)

        elif exc_info:
            self._task_create_stack()
            self._add_last_error_info(exc_info=exc_info)

        elif verbosity >= FormatterVerbosity.HIGH:
            self._task_create_stack()

    def add_log_info(self, task_run):
        logs = [("local", task_run.log.local_log_file)]
        if task_run.log.remote_log_file:
            logs.append(("remote", task_run.log.remote_log_file))
        self.banner.column("LOG", self.banner.f_simple_dict(logs))

    def add_external_resource_info(self, task_run):
        self.banner.column_properties(
            "EXTERNAL",
            [(k, v) for k, v in six.iteritems(task_run.external_resource_urls)],
        )

    def add_tracker_info(self, task_run):
        self.banner.column("TRACKER", task_run.task_tracker_url)

    def add_task_run_info(self, task_run):
        self.banner.column_properties(
            "TASK RUN",
            [
                ("task_run_uid", task_run.task_run_uid),
                ("task_run_attempt_uid", task_run.task_run_attempt_uid),
                ("state", task_run.task_run_state),
            ],
        )

    def add_task_info(self):
        from dbnd._core.task.task import DEFAULT_CLASS_VERSION

        task_params = [
            ("task_id", self.task.task_id),
            ("task_version", self.task.task_version),
        ]

        if self.task.task_class_version != DEFAULT_CLASS_VERSION:
            task_class_version = (
                _f_none_default(self.task.task_class_version, DEFAULT_CLASS_VERSION),
            )
            task_params.append(("task_class_version", task_class_version))

        task_params.extend(
            [
                ("env", self.task.task_env.task_name),
                ("env_cloud", self.task.task_env.cloud_type),
                ("env_label", _f_env(self.task.task_env.env_label)),
                ("task_target_date", self.task.task_target_date),
            ]
        )
        self.banner.column("TASK", self.banner.f_simple_dict(task_params))

    def add_disabled_info(self):
        disable = [
            ("task_enabled", self.task.task_enabled),
            ("task_enabled_in_prod", self.task.task_enabled_in_prod),
        ]

        self.banner.column("DISABLED", self.banner.f_simple_dict(disable))

    def add_time_info(self, task_run):
        time_fields = [("start", "%s" % task_run.start_time)]

        if task_run.finished_time:
            time_fields.append(("finished", "%s" % task_run.finished_time))
            task_duration = task_run.finished_time - task_run.start_time
            time_fields.append(("duration", "%s" % task_duration))

        self.banner.column("TIME", self.banner.f_simple_dict(time_fields))

    def _task_create_stack(self):
        task_call_source = task_call_source_to_str(self.task.task_meta.task_call_source)
        if task_call_source and not self.is_driver_or_submitter:
            self.banner.column("CREATED BY", value=task_call_source)

    def _add_last_error_info(self, exc_info=None):
        ex = exc_info[1]

        if show_exc_info(ex):
            # if we are showing internal task - we don't know how to "minimize" stack.
            # everything is databand.* ...
            isolate = not self.task.task_definition.full_task_family.startswith(
                "databand."
            )
            traceback_str = self.settings.log.format_exception_as_str(
                exc_info=exc_info, isolate=isolate
            )

            self.banner.column(
                colored("TRACEBACK", color="red", attrs=["bold"]),
                traceback_str,
                raw_name=True,
            )

        self.banner.column(
            colored("ERROR MESSAGE", color="red", attrs=["bold"]),
            str(ex),
            raw_name=True,
        )

        self.banner.column(
            colored("HELP", attrs=["bold"]),
            get_help_msg(ex),
            raw_name=True,
            skip_if_empty=True,
        )
        self.banner.column(
            colored("CAUSED BY", color="red", attrs=["bold"]),
            nested_exceptions_str(ex, limit=3),
            raw_name=True,
            skip_if_empty=True,
        )

    def _add_verbose_info(self):
        self.banner.new_section()
        self.banner.column(
            "INPUTS USER", self.banner.f_io(self.relations.task_inputs_user)
        )
        self.banner.column(
            "INPUTS SYSTEM", self.banner.f_struct(self.relations.task_inputs_system)
        )
        self.banner.new_section()

        self.banner.column("TASK_BAND", self.task.task_band)
        self.banner.column(
            "OUTPUTS USER", self.banner.f_io(self.relations.task_outputs_user)
        )
        self.banner.column(
            "OUTPUTS SYSTEM", self.banner.f_struct(self.relations.task_outputs_system)
        )
        self.banner.write("\n")

        self.banner.column("SIGNATURE SOURCE", self.task_meta.task_signature_source)
        self.banner.column(
            "TASK OUTPUTS SIGNATURE SOURCE",
            self.task_meta.task_outputs_signature_source,
        )

    def _add_task_band_info(self):
        self.banner.new_section()
        self.banner.column("TASK_BAND", self.task.task_band)


## Building Params Table ##


class _ParamTableDirector(object):
    """
    Describes how to build the param table
    """

    def __init__(self, task, banner):
        self.task = task
        self.banner = banner
        self.record_builder = _ParamRecordBuilder()
        self.header_builder = _ParamHeaderBuilder()

    def add_params_table(
        self,
        all_params=False,
        param_kind=True,
        param_type=True,
        param_format=False,
        param_source=False,
        param_section=False,
        param_default=False,
    ):
        # config
        self.kind = param_kind
        self.type = param_type
        self.format = param_format
        self.source = param_source
        self.section = param_section
        self.default = param_default

        """adds the param table to the banner"""
        from dbnd._core.decorator.schemed_result import ResultProxyTarget

        exclude = set(_TASK_FIELDS)  # excluding those

        # special treatment for result
        t_result = getattr(self.task, "result", None)

        if isinstance(t_result, ResultProxyTarget):
            self.banner.column("RESULT", t_result)
            exclude.add("result")

        params_data = []
        params_warnings = []
        # iterating over the params and building the data for the table
        for param_def, param_value, param_meta in self.task._params:
            # filter switches
            if not all_params:
                if param_def.name in exclude:
                    continue
                if param_def.system:
                    continue

            # building a single row
            param_row = self.build_record(param_def, param_meta, param_value)
            params_data.append(param_row)

            # extract param warnings
            if param_meta and param_meta.warnings:
                params_warnings.extend(param_meta.warnings)

        if params_warnings:
            self.banner.column("BUILD WARNINGS:", "")
            self.banner.write("".join("\t%s\n" % pw for pw in params_warnings))

        if params_data:
            params_header = self.build_headers()
            params_table = safe_tabulate(
                tabular_data=params_data, headers=params_header
            )
            self.banner.column("PARAMS:", "")
            self.banner.write(params_table)

        self.banner.new_line()

    def build_record(self, p, param_meta, value):
        self.record_builder.reset(self.task, p, value, param_meta)
        return self._build_row(self.record_builder)

    def build_headers(self):
        self.header_builder.reset()
        return self._build_row(self.header_builder)

    def _build_row(self, builder):
        """instructs the builder which of the columns to build"""
        builder.add_name()
        if self.kind:
            builder.add_kind()
        if self.type:
            builder.add_type()
        if self.format:
            builder.add_format()
        if self.source:
            builder.add_source()
        if self.default:
            builder.add_default()
        if self.section:
            builder.add_section()

        builder.add_value()

        return builder.row


class _ParamRecordBuilder(object):
    """Describe how to build each of the columns in the param table, from single param"""

    def __init__(self):
        self.task = None
        self.definition = None
        self.value = None
        self.meta = None
        self.row = None

    def reset(self, task, definition, value, meta):
        self.task = task
        self.definition = definition
        self.value = value if not definition.hidden else "***"
        self.meta = meta
        self.row = []

    def add_name(self):
        self.row.append(self.definition.name)

    def add_kind(self):
        p_kind = self._param_kind()
        self.row.append(p_kind)

    def _param_kind(self):
        if self.definition.is_output():
            p_kind = "output"
        elif not self.definition.load_on_build:
            p_kind = "input"
        else:
            p_kind = "param"
        return p_kind

    def add_type(self):
        self.row.append(self.definition.value_type_str)

    def add_source(self):
        source = self.meta.source if self.meta else ""
        self.row.append(source)

    def add_format(self):
        target_config = self.definition.target_config

        if isinstance(self.value, InMemoryTarget):
            target_config = "memory"

        if isinstance(self.value, DataTarget) and hasattr(self.value, "config"):
            target_config = self.value.config

        self.row.append(target_config)

    def add_section(self):
        section = self.definition.parameter_origin.get_task_family()
        if section == self.task.get_task_family():
            section = ""

        self.row.append(section)

    def add_default(self):
        self.row.append(str(self.definition.default))

    def add_value(self):
        console_value_preview_size = DescribeConfig.current().console_value_preview_size

        value_str = (
            TextBanner.f_io(self.value)
            if self._param_kind() in ["input", "output"]
            else self.definition.to_str(self.value)
        )

        value_str = safe_string(value_str, console_value_preview_size)

        # add preview
        if isinstance(self.value, Target) and self.value.target_meta:
            preview_value = safe_string(
                self.value.target_meta.value_preview, console_value_preview_size,
            )
            # we should add minimal preview
            if len(preview_value) < 100:
                value_str += " :='%s'" % preview_value

        if value_str and "\n" in value_str:
            # some simple heuristics around value
            extra_padding = " " * len("\t".join(map(str, self.row)))
            value_str = "".join(
                "\n%s%s" % (extra_padding, l) for l in value_str.split("\n")
            )
            value_str = "-->\n" + value_str

        self.row.append(value_str)


class _ParamHeaderBuilder(object):
    """Describe how to build each of the columns in the param table header"""

    def __init__(self):
        self.row = []

    def reset(self):
        self.row = []

    def add_name(self):
        self.row.append("Name")

    def add_kind(self):
        self.row.append("Kind")

    def add_type(self):
        self.row.append("Type")

    def add_source(self):
        self.row.append("Source")

    def add_format(self):
        self.row.append("Format")

    def add_section(self):
        self.row.append("Section")

    def add_default(self):
        self.row.append("Default")

    def add_value(self):
        self.row.append("-= Value =-")



from dbnd._core.tracking.schemas.base import ApiObjectSchema
from dbnd._vendor.marshmallow import fields


class AlertDefsSchema(ApiObjectSchema):
    uid = fields.Str()
    original_uid = fields.Str()
    custom_name = fields.Str()
    custom_description = fields.Str()
    summary = fields.Str()

    severity = fields.Str()
    type = fields.Str()
    task_name = fields.Str()
    user_metric = fields.Str()
    operator = fields.Str()
    is_str_value = fields.Bool()
    value = fields.Str()
    advanced_json = fields.Str()

    created_at = fields.DateTime()
    scheduled_job_uid = fields.Str()
    scheduled_job_name = fields.Str(attribute="scheduled_job.name")
    job_name = fields.Str()
    job_id = fields.Int()
    airflow_instance_name = fields.Str(attribute="job.airflow_server_info.name")
    env = fields.Str(attribute="job.airflow_server_info.env")



from targets.pipes.base import (
    InputPipeProcessWrapper,
    IOPipeline,
    OutputPipeProcessWrapper,
)


class Bzip2Pipeline(IOPipeline):

    input = "bytes"
    output = "bytes"

    def pipe_reader(self, input_pipe):
        return InputPipeProcessWrapper(["bzcat"], input_pipe)

    def pipe_writer(self, output_pipe):
        return OutputPipeProcessWrapper(["bzip2"], output_pipe)



"""Implementation of HjsonEncoder
"""
from __future__ import absolute_import
import re
from operator import itemgetter
from decimal import Decimal
from .compat import u, unichr, binary_type, string_types, integer_types, PY3

from dbnd._vendor.hjson.decoder import PosInf

# This is required because u() will mangle the string and ur'' isn't valid
# python3 syntax
ESCAPE = re.compile(u'[\\x00-\\x1f\\\\"\\b\\f\\n\\r\\t\u2028\u2029\uffff]')
ESCAPE_ASCII = re.compile(r'([\\"]|[^\ -~])')
HAS_UTF8 = re.compile(r'[\x80-\xff]')
ESCAPE_DCT = {
    '\\': '\\\\',
    '"': '\\"',
    '\b': '\\b',
    '\f': '\\f',
    '\n': '\\n',
    '\r': '\\r',
    '\t': '\\t',
}
for i in range(0x20):
    #ESCAPE_DCT.setdefault(chr(i), '\\u{0:04x}'.format(i))
    ESCAPE_DCT.setdefault(chr(i), '\\u%04x' % (i,))
for i in [0x2028, 0x2029, 0xffff]:
    ESCAPE_DCT.setdefault(unichr(i), '\\u%04x' % (i,))

COMMONRANGE=u'\x7f-\x9f\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff'

# NEEDSESCAPE tests if the string can be written without escapes
NEEDSESCAPE = re.compile(u'[\\\"\x00-\x1f'+COMMONRANGE+']')
# NEEDSQUOTES tests if the string can be written as a quoteless string (like needsEscape but without \\ and \")
NEEDSQUOTES = re.compile(u'^\\s|^"|^\'|^#|^\\/\\*|^\\/\\/|^\\{|^\\}|^\\[|^\\]|^:|^,|\\s$|[\x00-\x1f'+COMMONRANGE+u']')
# NEEDSESCAPEML tests if the string can be written as a multiline string (like needsEscape but without \n, \r, \\, \", \t)
NEEDSESCAPEML = re.compile(u'\'\'\'|^[\\s]+$|[\x00-\x08\x0b\x0c\x0e-\x1f'+COMMONRANGE+u']')

WHITESPACE = ' \t\n\r'
STARTSWITHNUMBER = re.compile(r'^[\t ]*(-?(?:0|[1-9]\d*))(\.\d+)?([eE][-+]?\d+)?\s*((,|\]|\}|#|\/\/|\/\*).*)?$');
STARTSWITHKEYWORD = re.compile(r'^(true|false|null)\s*((,|\]|\}|#|\/\/|\/\*).*)?$');
NEEDSESCAPENAME = re.compile(r'[,\{\[\}\]\s:#"\']|\/\/|\/\*|'+"'''")

FLOAT_REPR = repr

def encode_basestring(s, _PY3=PY3, _q=u('"')):
    """Return a JSON representation of a Python string

    """
    if _PY3:
        if isinstance(s, binary_type):
            s = s.decode('utf-8')
    else:
        if isinstance(s, str) and HAS_UTF8.search(s) is not None:
            s = s.decode('utf-8')
    def replace(match):
        return ESCAPE_DCT[match.group(0)]
    return _q + ESCAPE.sub(replace, s) + _q


def encode_basestring_ascii(s, _PY3=PY3):
    """Return an ASCII-only JSON representation of a Python string

    """
    if _PY3:
        if isinstance(s, binary_type):
            s = s.decode('utf-8')
    else:
        if isinstance(s, str) and HAS_UTF8.search(s) is not None:
            s = s.decode('utf-8')
    def replace(match):
        s = match.group(0)
        try:
            return ESCAPE_DCT[s]
        except KeyError:
            n = ord(s)
            if n < 0x10000:
                #return '\\u{0:04x}'.format(n)
                return '\\u%04x' % (n,)
            else:
                # surrogate pair
                n -= 0x10000
                s1 = 0xd800 | ((n >> 10) & 0x3ff)
                s2 = 0xdc00 | (n & 0x3ff)
                #return '\\u{0:04x}\\u{1:04x}'.format(s1, s2)
                return '\\u%04x\\u%04x' % (s1, s2)
    return '"' + str(ESCAPE_ASCII.sub(replace, s)) + '"'


class HjsonEncoder(object):
    """Extensible JSON <http://json.org> encoder for Python data structures.

    Supports the following objects and types by default:

    +-------------------+---------------+
    | Python            | JSON          |
    +===================+===============+
    | dict, namedtuple  | object        |
    +-------------------+---------------+
    | list, tuple       | array         |
    +-------------------+---------------+
    | str, unicode      | string        |
    +-------------------+---------------+
    | int, long, float  | number        |
    +-------------------+---------------+
    | True              | true          |
    +-------------------+---------------+
    | False             | false         |
    +-------------------+---------------+
    | None              | null          |
    +-------------------+---------------+

    To extend this to recognize other objects, subclass and implement a
    ``.default()`` method with another method that returns a serializable
    object for ``o`` if possible, otherwise it should call the superclass
    implementation (to raise ``TypeError``).

    """

    def __init__(self, skipkeys=False, ensure_ascii=True,
                 check_circular=True, sort_keys=False,
                 indent='  ', encoding='utf-8', default=None,
                 use_decimal=True, namedtuple_as_object=True,
                 tuple_as_array=True, bigint_as_string=False,
                 item_sort_key=None, for_json=False,
                 int_as_string_bitcount=None):
        """Constructor for HjsonEncoder, with sensible defaults.

        If skipkeys is false, then it is a TypeError to attempt
        encoding of keys that are not str, int, long, float or None.  If
        skipkeys is True, such items are simply skipped.

        If ensure_ascii is true, the output is guaranteed to be str
        objects with all incoming unicode characters escaped.  If
        ensure_ascii is false, the output will be unicode object.

        If check_circular is true, then lists, dicts, and custom encoded
        objects will be checked for circular references during encoding to
        prevent an infinite recursion (which would cause an OverflowError).
        Otherwise, no such check takes place.

        If sort_keys is true, then the output of dictionaries will be
        sorted by key; this is useful for regression tests to ensure
        that JSON serializations can be compared on a day-to-day basis.

        If indent is a string, then JSON array elements and object members
        will be pretty-printed with a newline followed by that string repeated
        for each level of nesting.

        If specified, default is a function that gets called for objects
        that can't otherwise be serialized.  It should return a JSON encodable
        version of the object or raise a ``TypeError``.

        If encoding is not None, then all input strings will be
        transformed into unicode using that encoding prior to JSON-encoding.
        The default is UTF-8.

        If use_decimal is true (not the default), ``decimal.Decimal`` will
        be supported directly by the encoder. For the inverse, decode JSON
        with ``parse_float=decimal.Decimal``.

        If namedtuple_as_object is true (the default), objects with
        ``_asdict()`` methods will be encoded as JSON objects.

        If tuple_as_array is true (the default), tuple (and subclasses) will
        be encoded as JSON arrays.

        If bigint_as_string is true (not the default), ints 2**53 and higher
        or lower than -2**53 will be encoded as strings. This is to avoid the
        rounding that happens in Javascript otherwise.

        If int_as_string_bitcount is a positive number (n), then int of size
        greater than or equal to 2**n or lower than or equal to -2**n will be
        encoded as strings.

        If specified, item_sort_key is a callable used to sort the items in
        each dictionary. This is useful if you want to sort items other than
        in alphabetical order by key.

        If for_json is true (not the default), objects with a ``for_json()``
        method will use the return value of that method for encoding as JSON
        instead of the object.

        """

        self.skipkeys = skipkeys
        self.ensure_ascii = ensure_ascii
        self.check_circular = check_circular
        self.sort_keys = sort_keys
        self.use_decimal = use_decimal
        self.namedtuple_as_object = namedtuple_as_object
        self.tuple_as_array = tuple_as_array
        self.bigint_as_string = bigint_as_string
        self.item_sort_key = item_sort_key
        self.for_json = for_json
        self.int_as_string_bitcount = int_as_string_bitcount
        if indent is not None and not isinstance(indent, string_types):
            indent = indent * ' '
        elif indent is None:
            indent = '  '
        self.indent = indent
        if default is not None:
            self.default = default
        self.encoding = encoding

    def default(self, o):
        """Implement this method in a subclass such that it returns
        a serializable object for ``o``, or calls the base implementation
        (to raise a ``TypeError``).

        For example, to support arbitrary iterators, you could
        implement default like this::

            def default(self, o):
                try:
                    iterable = iter(o)
                except TypeError:
                    pass
                else:
                    return list(iterable)
                return HjsonEncoder.default(self, o)

        """
        raise TypeError(repr(o) + " is not JSON serializable")

    def encode(self, o):
        """Return a JSON string representation of a Python data structure.

        >>> from hjson import HjsonEncoder
        >>> HjsonEncoder().encode({"foo": ["bar", "baz"]})
        '{"foo": ["bar", "baz"]}'

        """
        # This is for extremely simple cases and benchmarks.
        if isinstance(o, binary_type):
            _encoding = self.encoding
            if (_encoding is not None and not (_encoding == 'utf-8')):
                o = o.decode(_encoding)

        # This doesn't pass the iterator directly to ''.join() because the
        # exceptions aren't as detailed.  The list call should be roughly
        # equivalent to the PySequence_Fast that ''.join() would do.
        chunks = self.iterencode(o, _one_shot=True)
        if not isinstance(chunks, (list, tuple)):
            chunks = list(chunks)
        if self.ensure_ascii:
            return ''.join(chunks)
        else:
            return u''.join(chunks)

    def iterencode(self, o, _one_shot=False):
        """Encode the given object and yield each string
        representation as available.

        For example::

            for chunk in HjsonEncoder().iterencode(bigobject):
                mysocket.write(chunk)

        """
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = encode_basestring_ascii
        else:
            _encoder = encode_basestring
        if self.encoding != 'utf-8':
            def _encoder(o, _orig_encoder=_encoder, _encoding=self.encoding):
                if isinstance(o, binary_type):
                    o = o.decode(_encoding)
                return _orig_encoder(o)

        def floatstr(o, _repr=FLOAT_REPR, _inf=PosInf, _neginf=-PosInf):
            # Check for specials. Note that this type of test is processor
            # and/or platform-specific, so do tests which don't depend on
            # the internals.

            if o != o or o == _inf or o == _neginf:
                return 'null'
            else:
                return _repr(o)

        key_memo = {}
        int_as_string_bitcount = (
            53 if self.bigint_as_string else self.int_as_string_bitcount)
        _iterencode = _make_iterencode(
            markers, self.default, _encoder, self.indent, floatstr,
            self.sort_keys, self.skipkeys, _one_shot, self.use_decimal,
            self.namedtuple_as_object, self.tuple_as_array,
            int_as_string_bitcount,
            self.item_sort_key, self.encoding, self.for_json,
            Decimal=Decimal)
        try:
            return _iterencode(o, 0, True)
        finally:
            key_memo.clear()


def _make_iterencode(markers, _default, _encoder, _indent, _floatstr,
        _sort_keys, _skipkeys, _one_shot,
        _use_decimal, _namedtuple_as_object, _tuple_as_array,
        _int_as_string_bitcount, _item_sort_key,
        _encoding,_for_json,
        ## HACK: hand-optimized bytecode; turn globals into locals
        _PY3=PY3,
        ValueError=ValueError,
        string_types=string_types,
        Decimal=Decimal,
        dict=dict,
        float=float,
        id=id,
        integer_types=integer_types,
        isinstance=isinstance,
        list=list,
        str=str,
        tuple=tuple,
    ):
    if _item_sort_key and not callable(_item_sort_key):
        raise TypeError("item_sort_key must be None or callable")
    elif _sort_keys and not _item_sort_key:
        _item_sort_key = itemgetter(0)

    if (_int_as_string_bitcount is not None and
        (_int_as_string_bitcount <= 0 or
         not isinstance(_int_as_string_bitcount, integer_types))):
        raise TypeError("int_as_string_bitcount must be a positive integer")

    def _encode_int(value):
        return str(value)

    def _stringify_key(key):
        if isinstance(key, string_types): # pragma: no cover
            pass
        elif isinstance(key, binary_type):
            key = key.decode(_encoding)
        elif isinstance(key, float):
            key = _floatstr(key)
        elif key is True:
            key = 'true'
        elif key is False:
            key = 'false'
        elif key is None:
            key = 'null'
        elif isinstance(key, integer_types):
            key = str(key)
        elif _use_decimal and isinstance(key, Decimal):
            key = str(key)
        elif _skipkeys:
            key = None
        else:
            raise TypeError("key " + repr(key) + " is not a string")
        return key

    def _encoder_key(name):
        if not name: return '""'

        # Check if we can insert this name without quotes
        if NEEDSESCAPENAME.search(name):
            return _encoder(name)
        else:
          # return without quotes
          return name

    def _encoder_str(str, _current_indent_level):
        if not str: return '""'

        # Check if we can insert this string without quotes
        # see hjson syntax (must not parse as true, false, null or number)

        first = str[0]
        isNumber = False
        if first == '-' or first >= '0' and first <= '9':
            isNumber = STARTSWITHNUMBER.match(str) is not None

        if (NEEDSQUOTES.search(str) or
            isNumber or
            STARTSWITHKEYWORD.match(str) is not None):

            # If the string contains no control characters, no quote characters, and no
            # backslash characters, then we can safely slap some quotes around it.
            # Otherwise we first check if the string can be expressed in multiline
            # format or we must replace the offending characters with safe escape
            # sequences.

            if not NEEDSESCAPE.search(str):
                return '"' + str + '"'
            elif not NEEDSESCAPEML.search(str):
                return _encoder_str_ml(str, _current_indent_level + 1)
            else:
                return _encoder(str)
        else:
            # return without quotes
            return str

    def _encoder_str_ml(str, _current_indent_level):

        a = str.replace('\r', '').split('\n')
        # gap += indent;

        if len(a) == 1:
            # The string contains only a single line. We still use the multiline
            # format as it avoids escaping the \ character (e.g. when used in a
            # regex).
            return "'''" + a[0] + "'''"
        else:
            gap = _indent * _current_indent_level
            res = '\n' + gap + "'''"
            for line in a:
                res += '\n'
                if line: res += gap + line
            return res + '\n' + gap + "'''"

    def _iterencode_dict(dct, _current_indent_level, _isRoot=False):
        if not dct:
            yield '{}'
            return
        if markers is not None:
            markerid = id(dct)
            if markerid in markers:
                raise ValueError("Circular reference detected")
            markers[markerid] = dct

        if not _isRoot:
            yield '\n' + (_indent * _current_indent_level)

        _current_indent_level += 1
        newline_indent = '\n' + (_indent * _current_indent_level)

        yield '{'

        if _PY3:
            iteritems = dct.items()
        else:
            iteritems = dct.iteritems()
        if _item_sort_key:
            items = []
            for k, v in dct.items():
                if not isinstance(k, string_types):
                    k = _stringify_key(k)
                    if k is None:
                        continue
                items.append((k, v))
            items.sort(key=_item_sort_key)
        else:
            items = iteritems
        for key, value in items:
            if not (_item_sort_key or isinstance(key, string_types)):
                key = _stringify_key(key)
                if key is None:
                    # _skipkeys must be True
                    continue

            yield newline_indent
            yield _encoder_key(key)

            first = True
            for chunk in _iterencode(value, _current_indent_level):
                if first:
                    first = False
                    if chunk[0 : 1] == '\n': yield ':'
                    else: yield ': '
                yield chunk

        if newline_indent is not None:
            _current_indent_level -= 1
            yield '\n' + (_indent * _current_indent_level)
        yield '}'
        if markers is not None:
            del markers[markerid]


    def _iterencode_list(lst, _current_indent_level, _isRoot=False):
        if not lst:
            yield '[]'
            return
        if markers is not None:
            markerid = id(lst)
            if markerid in markers:
                raise ValueError("Circular reference detected")
            markers[markerid] = lst

        if not _isRoot:
            yield '\n' + (_indent * _current_indent_level)

        _current_indent_level += 1
        newline_indent = '\n' + (_indent * _current_indent_level)
        yield '['

        for value in lst:
            yield newline_indent

            for chunk in _iterencode(value, _current_indent_level, True):
                yield chunk

        if newline_indent is not None:
            _current_indent_level -= 1
            yield '\n' + (_indent * _current_indent_level)
        yield ']'
        if markers is not None:
            del markers[markerid]


    def _iterencode(o, _current_indent_level, _isRoot=False):
        if (isinstance(o, string_types) or
            (_PY3 and isinstance(o, binary_type))):
            yield _encoder_str(o, _current_indent_level)
        elif o is None:
            yield 'null'
        elif o is True:
            yield 'true'
        elif o is False:
            yield 'false'
        elif isinstance(o, integer_types):
            yield _encode_int(o)
        elif isinstance(o, float):
            yield _floatstr(o)
        else:
            for_json = _for_json and getattr(o, 'for_json', None)
            if for_json and callable(for_json):
                for chunk in _iterencode(for_json(), _current_indent_level, _isRoot):
                    yield chunk
            elif isinstance(o, list):
                for chunk in _iterencode_list(o, _current_indent_level, _isRoot):
                    yield chunk
            else:
                _asdict = _namedtuple_as_object and getattr(o, '_asdict', None)
                if _asdict and callable(_asdict):
                    for chunk in _iterencode_dict(_asdict(), _current_indent_level, _isRoot):
                        yield chunk
                elif (_tuple_as_array and isinstance(o, tuple)):
                    for chunk in _iterencode_list(o, _current_indent_level, _isRoot):
                        yield chunk
                elif isinstance(o, dict):
                    for chunk in _iterencode_dict(o, _current_indent_level, _isRoot):
                        yield chunk
                elif _use_decimal and isinstance(o, Decimal):
                    yield str(o)
                else:
                    if markers is not None:
                        markerid = id(o)
                        if markerid in markers:
                            raise ValueError("Circular reference detected")
                        markers[markerid] = o
                    o = _default(o)
                    for chunk in _iterencode(o, _current_indent_level, _isRoot):
                        yield chunk
                    if markers is not None:
                        del markers[markerid]

    return _iterencode



import os

from distutils.cmd import Command


class BdistFatZip(Command):

    description = "create deps and project distribution files for spark_submit"
    user_options = [
        ("dist-dir=", "d", "Build deps into dir. [default: dist]"),
        ("package-dir=", "p", "The dir of the setup.py file. [default: .]"),
        ("requirements=", "r", "Install from the given requirements file."),
    ]

    def initialize_options(self):
        self.dist_dir = "dist"
        self.package_dir = "."
        self.requirements = None

    def finalize_options(self):
        if self.requirements is not None:
            assert os.path.exists(
                self.requirements
            ), "requirements file '{}' does not exist.".format(self.requirements)

    def run(self):
        from dbnd.tasks.py_distribution.build_distribution import (
            build_fat_requirements_py_zip_file,
        )
        from targets.target_config import folder
        from targets import target

        build_fat_requirements_py_zip_file(
            target(self.dist_dir, config=folder), self.package_dir, self.requirements
        )



import logging
import typing

from typing import Any, Dict, Tuple, Type

import attr
import six


logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from dbnd._core.task.task import Task


@attr.s
class TaskCallState(object):
    started = attr.ib(default=False)
    finished = attr.ib(default=False)
    result_saved = attr.ib(default=False)

    result = attr.ib(default=None)

    should_store_result = attr.ib(default=False)

    def start(self):
        self.started = True
        self.finished = False
        self.result = None

    def finish(self, result=None):
        self.finished = True
        if self.should_store_result:
            self.result_saved = True
            self.result = result


CALL_FAILURE_OBJ = object()


@attr.s
class FuncCall(object):
    task_cls = attr.ib()  # type: Type[Task]
    call_args = attr.ib()  # type:  Tuple[Any]
    call_kwargs = attr.ib()  # type:  Dict[str,Any]
    call_user_code = attr.ib()

    def invoke(self):
        # make sure to invoke only with original decorator spec
        spec_args = self.task_cls._conf__decorator_spec.args
        kwargs = {
            key: value
            for key, value in six.iteritems(self.call_kwargs)
            if key in spec_args
        }

        func = self.call_user_code
        return func(*self.call_args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import shlex
import subprocess
import sys

from functools import partial

import six

from dbnd._core.cli.cmd_execute import execute
from dbnd._core.cli.cmd_heartbeat import send_heartbeat
from dbnd._core.cli.cmd_project import project_init
from dbnd._core.cli.cmd_run import run
from dbnd._core.cli.cmd_show import show_configs, show_tasks
from dbnd._core.cli.cmd_tracker import tracker
from dbnd._core.cli.cmd_utils import ipython
from dbnd._core.configuration.environ_config import should_fix_pyspark_imports
from dbnd._core.context.bootstrap import (
    _dbnd_exception_handling,
    dbnd_bootstrap,
    fix_pyspark_imports,
)
from dbnd._core.failures import dbnd_handle_errors
from dbnd._core.log.config import configure_basic_logging
from dbnd._core.plugin.dbnd_plugins import pm
from dbnd._core.plugin.dbnd_plugins_mng import register_dbnd_plugins
from dbnd._core.utils.platform import windows_compatible_mode
from dbnd._vendor import click
from dbnd._vendor.click_didyoumean import DYMGroup
from dbnd.cli.cmd_airflow_sync import airflow_sync
from dbnd.cli.cmd_alerts import alerts
from dbnd.cli.cmd_scheduler_management import schedule


logger = logging.getLogger(__name__)


@click.group(cls=DYMGroup)
def cli():
    dbnd_bootstrap()

    from dbnd import config

    # if we are running from "dbnd" entrypoint, we probably do not need to load Scheduled DAG
    # this will prevent from every airflow command to access dbnd web api
    if config.getboolean("airflow", "auto_disable_scheduled_dags_load"):
        os.environ["DBND_DISABLE_SCHEDULED_DAGS_LOAD"] = "True"
    pass


# project
cli.add_command(project_init)

# run
cli.add_command(run)
cli.add_command(execute)
cli.add_command(ipython)

# show
cli.add_command(show_configs)
cli.add_command(show_tasks)

# tracker
cli.add_command(tracker)

# heartbeat sender
cli.add_command(send_heartbeat)

# clients for the web-api
cli.add_command(alerts)
cli.add_command(airflow_sync)
cli.add_command(schedule)


@dbnd_handle_errors(exit_on_error=False)
def dbnd_cmd(command, args):
    """
    Invokes the passed dbnd command with CLI args emulation.

    Parameters:
        command (str): the command to be invoked
        args (Union[list, str]): list with CLI args to be emulated (if str is passed, it will be splitted)
    Returns:
        str: result of command execution
    """
    assert command in cli.commands
    if isinstance(args, six.string_types) or isinstance(args, six.text_type):
        args = shlex.split(args, posix=not windows_compatible_mode)
    current_argv = sys.argv
    logger.info("Running dbnd run: %s", subprocess.list2cmdline(args))
    try:
        sys.argv = [sys.executable, "-m", "databand", "run"] + args
        dbnd_bootstrap()
        return cli.commands[command](args=args, standalone_mode=False)
    finally:
        sys.argv = current_argv


dbnd_run_cmd = partial(dbnd_cmd, "run")
dbnd_schedule_cmd = partial(dbnd_cmd, "schedule")


def dbnd_run_cmd_main(task, env=None, args=None):
    """
    A wrapper for dbnd_cmd_run with error handling
    """
    from dbnd import Task

    if isinstance(task, Task):
        task_str = task.get_full_task_family()
    else:
        task_str = task

    try:
        cmd_args = [task_str]
        if env is not None:
            cmd_args = cmd_args + ["--env", env]
        if args is not None:
            cmd_args = cmd_args + args
        return dbnd_run_cmd(cmd_args)
    except KeyboardInterrupt:
        logger.error("Keyboard interrupt, exiting...")
        sys.exit(1)
    except Exception as ex:
        from dbnd._core.failures import get_databand_error_mesage

        msg, code = get_databand_error_mesage(ex=ex, args=sys.argv[1:])
        logger.error("dbnd cmd run failed with error: {}".format(msg))
        if code is not None:
            sys.exit(code)


def main():
    _dbnd_exception_handling()
    configure_basic_logging(None)
    if should_fix_pyspark_imports():
        fix_pyspark_imports()
    register_dbnd_plugins()

    # adding all plugins cli commands
    for commands in pm.hook.dbnd_get_commands():
        for command in commands:
            cli.add_command(command)
    try:
        return cli(prog_name="dbnd")

    except KeyboardInterrupt:
        sys.exit(1)

    except Exception as ex:
        from dbnd._core.failures import get_databand_error_mesage

        msg, code = get_databand_error_mesage(ex=ex, args=sys.argv[1:])
        logger.error(msg)
        if code is not None:
            sys.exit(code)

    except SystemExit as ex:
        if ex.code == 0:
            # databricks can't handle for a reason exit with an exit code.
            exit()
        else:
            sys.exit(ex.code)


if __name__ == "__main__":
    main()
